"""Prepare to aspirate command request, result, and implementation models."""

from __future__ import annotations
from opentrons_shared_data.errors.exceptions import PipetteOverpressureError
from pydantic import BaseModel
from typing import TYPE_CHECKING, Optional, Type, Union
from typing_extensions import Literal

from .pipetting_common import (
    OverpressureError,
    PipetteIdMixin,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)
from ..errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from ..execution import PipettingHandler, GantryMover
    from ..resources import ModelUtils


PrepareToAspirateCommandType = Literal["prepareToAspirate"]


class PrepareToAspirateParams(PipetteIdMixin):
    """Parameters required to prepare a specific pipette for aspiration."""

    pass


class PrepareToAspirateResult(BaseModel):
    """Result data from execution of an PrepareToAspirate command."""

    pass


_ExecuteReturn = Union[
    SuccessData[PrepareToAspirateResult],
    DefinedErrorData[OverpressureError],
]


class PrepareToAspirateImplementation(
    AbstractCommandImpl[PrepareToAspirateParams, _ExecuteReturn]
):
    """Prepare for aspirate command implementation."""

    def __init__(
        self,
        pipetting: PipettingHandler,
        model_utils: ModelUtils,
        gantry_mover: GantryMover,
        **kwargs: object,
    ) -> None:
        self._pipetting_handler = pipetting
        self._model_utils = model_utils
        self._gantry_mover = gantry_mover

    async def execute(self, params: PrepareToAspirateParams) -> _ExecuteReturn:
        """Prepare the pipette to aspirate."""
        current_position = await self._gantry_mover.get_position(params.pipetteId)
        try:
            await self._pipetting_handler.prepare_for_aspirate(
                pipette_id=params.pipetteId,
            )
        except PipetteOverpressureError as e:
            return DefinedErrorData(
                public=OverpressureError(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    wrappedErrors=[
                        ErrorOccurrence.from_failed(
                            id=self._model_utils.generate_id(),
                            createdAt=self._model_utils.get_timestamp(),
                            error=e,
                        )
                    ],
                    errorInfo=(
                        {
                            "retryLocation": (
                                current_position.x,
                                current_position.y,
                                current_position.z,
                            )
                        }
                    ),
                ),
            )
        else:
            return SuccessData(
                public=PrepareToAspirateResult(),
            )


class PrepareToAspirate(
    BaseCommand[PrepareToAspirateParams, PrepareToAspirateResult, ErrorOccurrence]
):
    """Prepare for aspirate command model."""

    commandType: PrepareToAspirateCommandType = "prepareToAspirate"
    params: PrepareToAspirateParams
    result: Optional[PrepareToAspirateResult]

    _ImplementationCls: Type[
        PrepareToAspirateImplementation
    ] = PrepareToAspirateImplementation


class PrepareToAspirateCreate(BaseCommandCreate[PrepareToAspirateParams]):
    """Prepare for aspirate command creation request model."""

    commandType: PrepareToAspirateCommandType = "prepareToAspirate"
    params: PrepareToAspirateParams

    _CommandCls: Type[PrepareToAspirate] = PrepareToAspirate



"""Dispense command request, result, and implementation models."""
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Union
from typing_extensions import Literal

from opentrons_shared_data.errors.exceptions import PipetteOverpressureError

from pydantic import Field

from ..types import DeckPoint
from ..state.update_types import StateUpdate, CLEAR
from .pipetting_common import (
    PipetteIdMixin,
    DispenseVolumeMixin,
    FlowRateMixin,
    LiquidHandlingWellLocationMixin,
    BaseLiquidHandlingResult,
    DestinationPositionResult,
    OverpressureError,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)
from ..errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from ..execution import MovementHandler, PipettingHandler
    from ..resources import ModelUtils
    from ..state.state import StateView


DispenseCommandType = Literal["dispense"]


class DispenseParams(
    PipetteIdMixin, DispenseVolumeMixin, FlowRateMixin, LiquidHandlingWellLocationMixin
):
    """Payload required to dispense to a specific well."""

    pushOut: Optional[float] = Field(
        None,
        description="push the plunger a small amount farther than necessary for accurate low-volume dispensing",
    )


class DispenseResult(BaseLiquidHandlingResult, DestinationPositionResult):
    """Result data from the execution of a Dispense command."""

    pass


_ExecuteReturn = Union[
    SuccessData[DispenseResult],
    DefinedErrorData[OverpressureError],
]


class DispenseImplementation(AbstractCommandImpl[DispenseParams, _ExecuteReturn]):
    """Dispense command implementation."""

    def __init__(
        self,
        state_view: StateView,
        movement: MovementHandler,
        pipetting: PipettingHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._movement = movement
        self._pipetting = pipetting
        self._model_utils = model_utils

    async def execute(self, params: DispenseParams) -> _ExecuteReturn:
        """Move to and dispense to the requested well."""
        state_update = StateUpdate()
        well_location = params.wellLocation
        labware_id = params.labwareId
        well_name = params.wellName
        volume = params.volume

        # TODO(pbm, 10-15-24): call self._state_view.geometry.validate_dispense_volume_into_well()

        position = await self._movement.move_to_well(
            pipette_id=params.pipetteId,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
        )
        deck_point = DeckPoint.construct(x=position.x, y=position.y, z=position.z)
        state_update.set_pipette_location(
            pipette_id=params.pipetteId,
            new_labware_id=labware_id,
            new_well_name=well_name,
            new_deck_point=deck_point,
        )

        try:
            volume = await self._pipetting.dispense_in_place(
                pipette_id=params.pipetteId,
                volume=volume,
                flow_rate=params.flowRate,
                push_out=params.pushOut,
            )
        except PipetteOverpressureError as e:
            state_update.set_liquid_operated(
                labware_id=labware_id,
                well_name=well_name,
                volume_added=CLEAR,
            )
            return DefinedErrorData(
                public=OverpressureError(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    wrappedErrors=[
                        ErrorOccurrence.from_failed(
                            id=self._model_utils.generate_id(),
                            createdAt=self._model_utils.get_timestamp(),
                            error=e,
                        )
                    ],
                    errorInfo={"retryLocation": (position.x, position.y, position.z)},
                ),
                state_update=state_update,
            )
        else:
            state_update.set_liquid_operated(
                labware_id=labware_id,
                well_name=well_name,
                volume_added=volume,
            )
            return SuccessData(
                public=DispenseResult(volume=volume, position=deck_point),
                state_update=state_update,
            )


class Dispense(BaseCommand[DispenseParams, DispenseResult, OverpressureError]):
    """Dispense command model."""

    commandType: DispenseCommandType = "dispense"
    params: DispenseParams
    result: Optional[DispenseResult]

    _ImplementationCls: Type[DispenseImplementation] = DispenseImplementation


class DispenseCreate(BaseCommandCreate[DispenseParams]):
    """Create dispense command request model."""

    commandType: DispenseCommandType = "dispense"
    params: DispenseParams

    _CommandCls: Type[Dispense] = Dispense



from typing import Callable, Dict, TYPE_CHECKING, Any

from typing_extensions import Protocol, TypedDict

from opentrons_shared_data.protocol.types import (
    BlowoutParams,
    DelayParams,
    PipetteAccessParams,
    StandardLiquidHandlingParams,
    TouchTipParams,
    MoveToSlotParams,
    TemperatureParams,
    ModuleIDParams,
    MagneticModuleEngageParams,
    ThermocyclerRunProfileParams,
    ThermocyclerSetTargetBlockParams,
    MoveToWellParams,
)

if TYPE_CHECKING:
    from opentrons.protocol_api.protocol_context import ProtocolContext
    from opentrons.protocol_api.instrument_context import InstrumentContext
    from opentrons.protocol_api.module_contexts import (
        MagneticModuleContext,
        ThermocyclerContext,
        TemperatureModuleContext,
    )
    from opentrons.protocol_api.labware import Labware


# this file defines types that require dev dependencies
# and are only relevant for static typechecking.
#
#  - code should be written so that this file can fail to import
#  - or the things defined in here can be None at execution time
#  - only types that match the above criteria should be put here
#  - please include this file as close to a leaf as possible

if Protocol is not None:

    class Dictable(Protocol):
        """A protocol defining the _asdict interface to
        classes generated from typing.NamedTuple, which cannot
        be used as a type annotation because it is a type ctor
        not a type (https://github.com/python/mypy/issues/3915)
        """

        async def _asdict(self) -> Dict[Any, Any]:
            ...


# using a lot of string literals here instead of the enum values from
# opentrons_shared_data.protocol.constants because of
# https://github.com/python/mypy/issues/4128
PipetteDispatch = TypedDict(
    "PipetteDispatch",
    {
        "delay": Callable[["ProtocolContext", "DelayParams"], None],
        "blowout": Callable[
            [Dict[str, "InstrumentContext"], Dict[str, "Labware"], "BlowoutParams"],
            None,
        ],
        "pickUpTip": Callable[
            [
                Dict[str, "InstrumentContext"],
                Dict[str, "Labware"],
                "PipetteAccessParams",
            ],
            None,
        ],
        "dropTip": Callable[
            [
                Dict[str, "InstrumentContext"],
                Dict[str, "Labware"],
                "PipetteAccessParams",
            ],
            None,
        ],
        "aspirate": Callable[
            [
                Dict[str, "InstrumentContext"],
                Dict[str, "Labware"],
                "StandardLiquidHandlingParams",
            ],
            None,
        ],
        "dispense": Callable[
            [
                Dict[str, "InstrumentContext"],
                Dict[str, "Labware"],
                "StandardLiquidHandlingParams",
            ],
            None,
        ],
        "touchTip": Callable[
            [Dict[str, "InstrumentContext"], Dict[str, "Labware"], "TouchTipParams"],
            None,
        ],
        "moveToSlot": Callable[
            ["ProtocolContext", Dict[str, "InstrumentContext"], "MoveToSlotParams"],
            None,
        ],
        "moveToWell": Callable[
            [Dict[str, "InstrumentContext"], Dict[str, "Labware"], "MoveToWellParams"],
            None,
        ],
        "airGap": Callable[
            [
                Dict[str, "InstrumentContext"],
                Dict[str, "Labware"],
                "StandardLiquidHandlingParams",
            ],
            None,
        ],
    },
    total=False,
)


JsonV4MagneticModuleDispatch = TypedDict(
    "JsonV4MagneticModuleDispatch",
    {
        "magneticModule/engageMagnet": Callable[
            ["MagneticModuleContext", "MagneticModuleEngageParams"], None
        ],
        "magneticModule/disengageMagnet": Callable[
            ["MagneticModuleContext", "ModuleIDParams"], None
        ],
    },
)

JsonV4TemperatureModuleDispatch = TypedDict(
    "JsonV4TemperatureModuleDispatch",
    {
        "temperatureModule/setTargetTemperature": Callable[
            ["TemperatureModuleContext", "TemperatureParams"], None
        ],
        "temperatureModule/deactivate": Callable[
            ["TemperatureModuleContext", "ModuleIDParams"], None
        ],
        "temperatureModule/awaitTemperature": Callable[
            ["TemperatureModuleContext", "TemperatureParams"], None
        ],
    },
)

JsonV4ThermocyclerDispatch = TypedDict(
    "JsonV4ThermocyclerDispatch",
    {
        "thermocycler/closeLid": Callable[
            ["ThermocyclerContext", "ModuleIDParams"], None
        ],
        "thermocycler/openLid": Callable[
            ["ThermocyclerContext", "ModuleIDParams"], None
        ],
        "thermocycler/deactivateBlock": Callable[
            ["ThermocyclerContext", "ModuleIDParams"], None
        ],
        "thermocycler/deactivateLid": Callable[
            ["ThermocyclerContext", "ModuleIDParams"], None
        ],
        "thermocycler/setTargetBlockTemperature": Callable[
            ["ThermocyclerContext", "ThermocyclerSetTargetBlockParams"], None
        ],
        "thermocycler/setTargetLidTemperature": Callable[
            ["ThermocyclerContext", "TemperatureParams"], None
        ],
        "thermocycler/runProfile": Callable[
            ["ThermocyclerContext", "ThermocyclerRunProfileParams"], None
        ],
        "thermocycler/awaitBlockTemperature": Callable[
            ["ThermocyclerContext", "TemperatureParams"], None
        ],
        "thermocycler/awaitLidTemperature": Callable[
            ["ThermocyclerContext", "TemperatureParams"], None
        ],
        "thermocycler/awaitProfileComplete": Callable[
            ["ThermocyclerContext", "ModuleIDParams"], None
        ],
    },
)



from typing import Optional, Tuple, List, AsyncIterator, Union
import contextlib
from typing_extensions import Protocol

from opentrons import types as top_types
from opentrons.config.types import (
    CapacitivePassSettings,
)
from opentrons.hardware_control.types import (
    Axis,
    OT3Mount,
    InstrumentProbeType,
    GripperProbe,
)
from opentrons.hardware_control.instruments.ot3.instrument_calibration import (
    GripperCalibrationOffset,
    PipetteOffsetSummary,
)
from opentrons.hardware_control.modules.module_calibration import (
    ModuleCalibrationOffset,
)


class FlexCalibratable(Protocol):
    """Calibration extensions for Flex hardware."""

    async def capacitive_probe(
        self,
        mount: OT3Mount,
        moving_axis: Axis,
        target_pos: float,
        pass_settings: CapacitivePassSettings,
        retract_after: bool = True,
        probe: Optional[InstrumentProbeType] = None,
    ) -> Tuple[float, bool]:
        """Determine the position of something using the capacitive sensor.

        This function orchestrates detecting the position of a collision between the
        capacitive probe on the tool on the specified mount, and some fixed element
        of the robot.

        When calling this function, the mount's probe critical point should already
        be aligned in the probe axis with the item to be probed.

        It will move the mount's probe critical point to a small distance behind
        the expected position of the element (which is target_pos, in deck coordinates,
        in the axis to be probed) while running the tool's capacitive sensor. When the
        sensor senses contact, the mount stops.

        This function moves away and returns the sensed position.

        This sensed position can be used in several ways, including
        - To get an absolute position in deck coordinates of whatever was
        targeted, if something was guaranteed to be physically present.
        - To detect whether a collision occured at all. If this function
        returns a value far enough past the anticipated position, then it indicates
        there was no material there.
        """
        ...

    async def capacitive_sweep(
        self,
        mount: OT3Mount,
        moving_axis: Axis,
        begin: top_types.Point,
        end: top_types.Point,
        speed_mm_s: float,
    ) -> List[float]:
        ...

    # Note that there is a default implementation of this function to allow for
    # the asynccontextmanager decorator to propagate properly.
    @contextlib.asynccontextmanager
    async def restore_system_constrants(self) -> AsyncIterator[None]:
        yield

    async def set_system_constraints_for_calibration(self) -> None:
        ...

    async def reset_instrument_offset(
        self, mount: Union[top_types.Mount, OT3Mount], to_default: bool = True
    ) -> None:
        ...

    def add_gripper_probe(self, probe: GripperProbe) -> None:
        ...

    def remove_gripper_probe(self) -> None:
        ...

    async def save_instrument_offset(
        self, mount: Union[top_types.Mount, OT3Mount], delta: top_types.Point
    ) -> Union[GripperCalibrationOffset, PipetteOffsetSummary]:
        ...

    async def save_module_offset(
        self, module_id: str, mount: OT3Mount, slot: str, offset: top_types.Point
    ) -> Optional[ModuleCalibrationOffset]:
        ...



"""CSV Parameter definition and associated classes/functions."""
from typing import Optional

from opentrons.protocol_engine.types import (
    RunTimeParameter,
    CSVParameter as ProtocolEngineCSVParameter,
    FileInfo,
)
from opentrons.protocols.api_support.types import APIVersion

from . import validation
from .parameter_definition import AbstractParameterDefinition
from .csv_parameter_interface import CSVParameter


class CSVParameterDefinition(AbstractParameterDefinition[Optional[bytes]]):
    """The definition for a user defined CSV file parameter."""

    def __init__(
        self, display_name: str, variable_name: str, description: Optional[str]
    ) -> None:
        """Initializes a CSV file parameter.

        Arguments:
            display_name: The display name of the parameter as it would show up on the frontend.
            variable_name: The variable name the parameter will be referred to in the run context.
            description: An optional description for the parameter.
        """
        self._display_name = validation.ensure_display_name(display_name)
        self._variable_name = validation.ensure_variable_name(variable_name)
        self._description = validation.ensure_description(description)
        self._value: Optional[bytes] = None
        self._file_info: Optional[FileInfo] = None

    @property
    def variable_name(self) -> str:
        """The in-protocol variable name of the parameter."""
        return self._variable_name

    @property
    def value(self) -> Optional[bytes]:
        """The current set file for the CSV parameter. Defaults to None on definition creation."""
        return self._value

    @value.setter
    def value(self, contents: bytes) -> None:
        self._value = contents

    @property
    def file_info(self) -> Optional[FileInfo]:
        return self._file_info

    @file_info.setter
    def file_info(self, file_info: FileInfo) -> None:
        self._file_info = file_info

    def as_csv_parameter_interface(self, api_version: APIVersion) -> CSVParameter:
        return CSVParameter(contents=self._value, api_version=api_version)

    def as_protocol_engine_type(self) -> RunTimeParameter:
        """Returns CSV parameter as a Protocol Engine type to send to client."""
        return ProtocolEngineCSVParameter(
            displayName=self._display_name,
            variableName=self._variable_name,
            description=self._description,
            file=self._file_info,
        )


def create_csv_parameter(
    display_name: str,
    variable_name: str,
    description: Optional[str] = None,
) -> CSVParameterDefinition:
    return CSVParameterDefinition(
        display_name=display_name, variable_name=variable_name, description=description
    )



"""Heater-Shaker Module sub-state."""
from dataclasses import dataclass
from typing import List, NewType, Optional, Dict

from opentrons.protocol_engine.errors import CannotPerformModuleAction

AbsorbanceReaderId = NewType("AbsorbanceReaderId", str)
AbsorbanceReaderLidId = NewType("AbsorbanceReaderLidId", str)
AbsorbanceReaderMeasureMode = NewType("AbsorbanceReaderMeasureMode", str)


# todo(mm, 2024-11-08): frozen=True is getting pretty painful because ModuleStore has
# no type-safe way to modify just a single attribute. Consider unfreezing this
# (taking care to ensure that consumers of ModuleView still only get a read-only view).
@dataclass(frozen=True)
class AbsorbanceReaderSubState:
    """Absorbance-Plate-Reader-specific state."""

    module_id: AbsorbanceReaderId
    configured: bool
    measured: bool
    is_lid_on: bool
    data: Optional[Dict[int, Dict[str, float]]]
    configured_wavelengths: Optional[List[int]]
    measure_mode: Optional[AbsorbanceReaderMeasureMode]
    reference_wavelength: Optional[int]

    def raise_if_lid_status_not_expected(self, lid_on_expected: bool) -> None:
        """Raise if the lid status is not correct."""
        match = self.is_lid_on is lid_on_expected
        if not match:
            raise CannotPerformModuleAction(
                "Cannot perform lid action because the lid is already "
                f"{'closed' if self.is_lid_on else 'open'}"
            )



class SubsystemUpdating(RuntimeError):
    def __init__(self, msg: str) -> None:
        self.msg = msg

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.msg}>"

    def __str__(self) -> str:
        return self.msg



"""Move to addressable area for drop tip command request, result, and implementation models."""
from __future__ import annotations
from pydantic import Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from ..errors import LocationNotAccessibleByPipetteError
from ..state import update_types
from ..types import DeckPoint, AddressableOffsetVector
from ..resources import fixture_validation
from .pipetting_common import (
    PipetteIdMixin,
    MovementMixin,
    DestinationPositionResult,
)
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from ..execution import MovementHandler
    from ..state.state import StateView

MoveToAddressableAreaForDropTipCommandType = Literal["moveToAddressableAreaForDropTip"]


class MoveToAddressableAreaForDropTipParams(PipetteIdMixin, MovementMixin):
    """Payload required to move a pipette to a specific addressable area.

    An *addressable area* is a space in the robot that may or may not be usable depending on how
    the robot's deck is configured. For example, if a Flex is configured with a waste chute, it will
    have additional addressable areas representing the opening of the waste chute, where tips and
    labware can be dropped.

    This moves the pipette so all of its nozzles are centered over the addressable area.
    If the pipette is currently configured with a partial tip layout, this centering is over all
    the pipette's physical nozzles, not just the nozzles that are active.

    The z-position will be chosen to put the bottom of the tips---or the bottom of the nozzles,
    if there are no tips---level with the top of the addressable area.

    When this command is executed, Protocol Engine will make sure the robot's deck is configured
    such that the requested addressable area actually exists. For example, if you request
    the addressable area B4, it will make sure the robot is set up with a B3/B4 staging area slot.
    If that's not the case, the command will fail.
    """

    addressableAreaName: str = Field(
        ...,
        description=(
            "The name of the addressable area that you want to use."
            " Valid values are the `id`s of `addressableArea`s in the"
            " [deck definition](https://github.com/Opentrons/opentrons/tree/edge/shared-data/deck)."
        ),
    )
    offset: AddressableOffsetVector = Field(
        AddressableOffsetVector(x=0, y=0, z=0),
        description="Relative offset of addressable area to move pipette's critical point.",
    )
    alternateDropLocation: Optional[bool] = Field(
        False,
        description=(
            "Whether to alternate location where tip is dropped within the addressable area."
            " If True, this command will ignore the offset provided and alternate"
            " between dropping tips at two predetermined locations inside the specified"
            " labware well."
            " If False, the tip will be dropped at the top center of the area."
        ),
    )
    ignoreTipConfiguration: Optional[bool] = Field(
        True,
        description=(
            "Whether to utilize the critical point of the tip configuraiton when moving to an addressable area."
            " If True, this command will ignore the tip configuration and use the center of the entire instrument"
            " as the critical point for movement."
            " If False, this command will use the critical point provided by the current tip configuration."
        ),
    )


class MoveToAddressableAreaForDropTipResult(DestinationPositionResult):
    """Result data from the execution of a MoveToAddressableAreaForDropTip command."""

    pass


class MoveToAddressableAreaForDropTipImplementation(
    AbstractCommandImpl[
        MoveToAddressableAreaForDropTipParams,
        SuccessData[MoveToAddressableAreaForDropTipResult],
    ]
):
    """Move to addressable area for drop tip command implementation."""

    def __init__(
        self, movement: MovementHandler, state_view: StateView, **kwargs: object
    ) -> None:
        self._movement = movement
        self._state_view = state_view

    async def execute(
        self, params: MoveToAddressableAreaForDropTipParams
    ) -> SuccessData[MoveToAddressableAreaForDropTipResult]:
        """Move the requested pipette to the requested addressable area in preperation of a drop tip."""
        state_update = update_types.StateUpdate()

        self._state_view.addressable_areas.raise_if_area_not_in_deck_configuration(
            params.addressableAreaName
        )

        if fixture_validation.is_staging_slot(params.addressableAreaName):
            raise LocationNotAccessibleByPipetteError(
                f"Cannot move pipette to staging slot {params.addressableAreaName}"
            )

        if params.alternateDropLocation:
            offset = self._state_view.geometry.get_next_tip_drop_location_for_addressable_area(
                addressable_area_name=params.addressableAreaName,
                pipette_id=params.pipetteId,
            )
        else:
            offset = params.offset

        x, y, z = await self._movement.move_to_addressable_area(
            pipette_id=params.pipetteId,
            addressable_area_name=params.addressableAreaName,
            offset=offset,
            force_direct=params.forceDirect,
            minimum_z_height=params.minimumZHeight,
            speed=params.speed,
            ignore_tip_configuration=params.ignoreTipConfiguration,
        )
        deck_point = DeckPoint.construct(x=x, y=y, z=z)
        state_update.set_pipette_location(
            pipette_id=params.pipetteId,
            new_addressable_area_name=params.addressableAreaName,
            new_deck_point=deck_point,
        )

        return SuccessData(
            public=MoveToAddressableAreaForDropTipResult(
                position=DeckPoint(x=x, y=y, z=z)
            ),
            state_update=state_update,
        )


class MoveToAddressableAreaForDropTip(
    BaseCommand[
        MoveToAddressableAreaForDropTipParams,
        MoveToAddressableAreaForDropTipResult,
        ErrorOccurrence,
    ]
):
    """Move to addressable area for drop tip command model."""

    commandType: MoveToAddressableAreaForDropTipCommandType = (
        "moveToAddressableAreaForDropTip"
    )
    params: MoveToAddressableAreaForDropTipParams
    result: Optional[MoveToAddressableAreaForDropTipResult]

    _ImplementationCls: Type[
        MoveToAddressableAreaForDropTipImplementation
    ] = MoveToAddressableAreaForDropTipImplementation


class MoveToAddressableAreaForDropTipCreate(
    BaseCommandCreate[MoveToAddressableAreaForDropTipParams]
):
    """Move to addressable area for drop tip command creation request model."""

    commandType: MoveToAddressableAreaForDropTipCommandType = (
        "moveToAddressableAreaForDropTip"
    )
    params: MoveToAddressableAreaForDropTipParams

    _CommandCls: Type[MoveToAddressableAreaForDropTip] = MoveToAddressableAreaForDropTip



from typing import Optional


# TODO(mc, 2022-10-20): move to public ProtocolContext
# once `Deck` and `ProtocolEngine` play nicely together
class InvalidModuleLocationError(ValueError):
    """Error raised if a load location for a module is invalid."""

    def __init__(self, invalid_value: Optional[str], module_name: str) -> None:
        """Initialize the error and message with the invalid value."""
        super().__init__(
            f"{invalid_value} is not a valid load location for {module_name}."
        )
        self.invalid_value = invalid_value
        self.module_name = module_name


class InvalidMagnetEngageHeightError(ValueError):
    """Error raised if a Magnetic Module engage height is invalid."""



from typing import Dict, Optional
from opentrons.util.async_helpers import ensure_yield
from opentrons.drivers.heater_shaker.abstract import AbstractHeaterShakerDriver
from opentrons.drivers.types import Temperature, RPM, HeaterShakerLabwareLatchStatus


class SimulatingDriver(AbstractHeaterShakerDriver):
    DEFAULT_TEMP = 23

    def __init__(self, serial_number: Optional[str] = None) -> None:
        self._labware_latch_state = HeaterShakerLabwareLatchStatus.IDLE_UNKNOWN
        self._current_temperature = self.DEFAULT_TEMP
        self._temperature = Temperature(current=self.DEFAULT_TEMP, target=None)
        self._rpm = RPM(current=0, target=None)
        self._homing_status = True
        self._serial_number = serial_number

    @ensure_yield
    async def connect(self) -> None:
        pass

    @ensure_yield
    async def disconnect(self) -> None:
        pass

    @ensure_yield
    async def is_connected(self) -> bool:
        return True

    @ensure_yield
    async def open_labware_latch(self) -> None:
        self._labware_latch_state = HeaterShakerLabwareLatchStatus.IDLE_OPEN

    @ensure_yield
    async def close_labware_latch(self) -> None:
        self._labware_latch_state = HeaterShakerLabwareLatchStatus.IDLE_CLOSED

    @ensure_yield
    async def get_labware_latch_status(self) -> HeaterShakerLabwareLatchStatus:
        return self._labware_latch_state

    @ensure_yield
    async def set_temperature(self, temperature: float) -> None:
        self._temperature.current = temperature
        self._temperature.target = temperature if temperature != 0.0 else None

    @ensure_yield
    async def get_temperature(self) -> Temperature:
        return self._temperature

    @ensure_yield
    async def set_rpm(self, rpm: int) -> None:
        self._rpm.target = rpm if rpm != 0 else None
        self._rpm.current = rpm
        self._homing_status = False

    @ensure_yield
    async def get_rpm(self) -> RPM:
        return self._rpm

    @ensure_yield
    async def home(self) -> None:
        self._rpm.target = None
        self._rpm.current = 0
        self._homing_status = True

    @ensure_yield
    async def deactivate_heater(self) -> None:
        self._temperature.target = None
        self._temperature.current = 23

    @ensure_yield
    async def deactivate_shaker(self) -> None:
        self._rpm.target = 0
        self._rpm.current = 0

    @ensure_yield
    async def deactivate(self) -> None:
        self._temperature.target = None
        self._temperature.current = 23
        self._rpm.target = 0
        self._rpm.current = 0

    @ensure_yield
    async def get_device_info(self) -> Dict[str, str]:
        return {
            "serial": self._serial_number if self._serial_number else "dummySerialHS",
            "model": "dummyModelHS",
            "version": "dummyVersionHS",
        }

    @ensure_yield
    async def enter_programming_mode(self) -> None:
        pass



from __future__ import annotations
import enum
import re
from itertools import groupby
from dataclasses import dataclass
from typing import List, Optional, Tuple
from opentrons.hardware_control.types import BoardRevision


class PinDir(enum.Enum):
    rev_input = enum.auto()
    input = enum.auto()
    output = enum.auto()


class PortGroup:
    MAIN = "main"
    LEFT = "left"
    RIGHT = "right"
    FRONT = "front"
    UNKNOWN = "unknown"


REV_OG_USB_PORTS = {"3": 1, "5": 2}
REV_A_USB_HUB = 3
FLEX_B2_USB_PORT_GROUP_LEFT = 4
FLEX_B2_USB_PORT_GROUP_RIGHT = 3
FLEX_B2_USB_PORT_GROUP_FRONT = 7
FLEX_B2_USB_PORTS = {"4": 1, "3": 2, "2": 3, "1": 4}

BUS_PATH = "/sys/bus/usb/devices/usb1/"

# Example usb path might look like:
# '/sys/bus/usb/devices/usb1/1-1/1-1.3/1-1.3:1.0/tty/ttyACM1/dev'.
# Example hid device path might look like:
# '/sys/bus/usb/devices/usb1/1-1/1-1.3/1-1.3:1.0/0003:16D0:1199.0002/hidraw/hidraw0/dev'
# There is only 1 bus that supports USB on the raspberry pi.
USB_PORT_INFO = re.compile(
    r"""
    (?P<port_path>(\d[\d]*-\d[\d\.]*[/]?)+):
    (?P<device_path>
        \d.\d/
        (tty/tty(\w{4})/dev | [\w:\.]+?/hidraw/hidraw\d/dev)
    )
    """,
    re.VERBOSE,
)

HUB_PATTERN = re.compile(r"(\d-[\d.]+\d?)[\/:]")


@dataclass(frozen=True)
class USBPort:
    name: str
    port_number: int
    port_group: str = PortGroup.UNKNOWN
    hub: bool = False
    hub_port: Optional[int] = None
    device_path: str = ""

    @classmethod
    def build(
        cls, full_path: str, board_revision: BoardRevision
    ) -> Optional["USBPort"]:
        """
        Build a USBPort dataclass.

        An example port path:
        `1-1.3/1-1.3:1.0/tty/ttyACM1/dev`

        Args:
            port_path: Full path of a usb device
            board_revision: Board revision

        Returns:
            Tuple of the port number, port group, hub, hub port, device path, and name
        """
        match = USB_PORT_INFO.search(full_path)
        if not match:
            return None

        port_path = match.group("port_path")
        device_path = match.group("device_path")
        port_nodes = cls.get_unique_nodes(port_path)
        hub, port, hub_port, name = cls.find_hub(port_nodes, board_revision)
        hub, port_group, port, hub_port = cls.map_to_revision(
            board_revision,
            (
                hub,
                port,
                hub_port,
            ),
        )
        return cls(
            name=name,
            port_number=port,
            port_group=port_group,
            hub=hub,
            hub_port=hub_port,
            device_path=device_path,
        )

    @staticmethod
    def find_hub(
        port_nodes: List[str],
        board_revision: BoardRevision,
    ) -> Tuple[Optional[int], int, Optional[int], str]:
        """
        Find the hub, port, and hub_port data for a USB port from
        the port nodes.

        The last item of the port nodes list is parsed by splitting
        the item at each period. It is parsed as follows:
        'bus.hub.port.hub_port'. The USB bus data is unused. The
        hub_port data is only populated if a USB hub is connected
        to that port. The Flex FRONT USB port is parsed as:
        'bus.hub.hub_port'.

        :param port_nodes: A list of unique port id(s)
        :returns: Tuple of the hub, port number, hub_port and name
        """
        if len(port_nodes) == 1 and "." not in port_nodes[0]:
            # if no hub is attached, such as on a dev kit, the port
            # nodes available will be 1-1
            port = int(port_nodes[0].split("-")[1])
            hub = None
            hub_port = None
            name = port_nodes[0]
        else:
            port_nodes = [node for node in port_nodes if "." in node]
            if len(port_nodes) > 2:
                port_info = port_nodes[2].split(".")
                hub = int(port_info[1])
                port = int(port_info[2])
                hub_port = int(port_info[3])
                name = port_nodes[2]
            elif len(port_nodes) > 1:
                if board_revision == BoardRevision.OG:
                    port_info = port_nodes[1].split(".")
                    hub = int(port_info[1])
                    port = int(port_info[1])
                    hub_port = int(port_info[2])
                    name = port_nodes[1]
                else:
                    port_info = port_nodes[1].split(".")
                    hub = int(port_info[1])
                    name = port_nodes[1]
                    if (board_revision == BoardRevision.FLEX_B2) and (
                        hub == FLEX_B2_USB_PORT_GROUP_FRONT
                    ):
                        port = 9
                        hub_port = int(port_info[2])
                    else:
                        port = int(port_info[2])
                        hub_port = None
            else:
                if board_revision == BoardRevision.FLEX_B2:
                    port_info = port_nodes[0].split(".")
                    hub = int(port_info[1])
                    port = 9
                    hub_port = None
                    name = port_nodes[0]
                else:
                    port = int(port_nodes[0].split(".")[1])
                    hub = None
                    hub_port = None
                    name = port_nodes[0]
        return hub, port, hub_port, name

    @staticmethod
    def get_unique_nodes(full_name: str) -> List[str]:
        """
        Get unique port nodes for a USB port from the USB port path.

        For a Flex or OT-2_R with a USB hub connected to a port,
        the USB port path will look like:
        `1-1.3/1-1.3.2/1-1.3.2.4/1-1.3.2.4`. This will become the
        following 3 port nodes: ['1-1.3', '1-1.3.2', '1-1.3.2.4'].
        The Flex FRONT USB port with a hub connected will look like:
        `1-1.7/1-1.7.4/1-1.7.4` and become ['1-1.7', '1-1.7.4'].

        For a Flex or OT-2_R without a USB hub connected to a port,
        the USB port path will look like: `1-1.3/1-1.3.4/1-1.3.4`.
        This will become the follwing 2 port nodes: ['1-1.3', '1-1.3.4'].
        The Flex FRONT USB port without a hub connected will look like:
        `1-1.7/1-1.7` and become ['1-1.7'].

        For a OT-2_OG with a USB hub connected to a port, the USB
        port path will look like: `1-1.3/1-1.3/1-1.3.3/1-1.3.3`. This will
        become the following 2 port nodes: ['1-1.3', '1-1.3.3'].

        For a OT-2_OG without a USB hub connected to a port, the USB
        port path will look like: `1-1.3/1-1.3`. This will become the
        following 1 port node: ['1-1.3'].

        We only need one unique id here, so we will filter out duplicates.

        :param full_name: Full path of the physical USB Path
        :returns: List of separated USB port paths
        """
        all_match = HUB_PATTERN.findall(full_name)
        match_set = []
        for match in all_match:
            if match not in match_set:
                match_set.append(match)
        return match_set

    @staticmethod
    def map_to_revision(
        board_revision: BoardRevision,
        port_info: Tuple[Optional[int], int, Optional[int]],
    ) -> Tuple[bool, str, int, Optional[int]]:
        """
        Synthesize the hub, port_group, and hub_port data for a USB port
        from the hub and hub_port data.

        If a USB hub is connected to a port, the hub data field is True
        and the hub_port data field is populated. Otherwise, the hub
        data field is False and the hub_port data field is None.

        For the OT-2, there is only one bank of USB ports, so the
        port_group is always MAIN. For the Flex, there are LEFT, RIGHT,
        and FRONT USB port banks, which map to specific hub values.

        For the Flex, the RIGHT port values are increased by 4 to match
        the physical hardware labeling (USB5 - USB8).

        :param port_info: Tuple of the hub, port number, and hub port
        :returns: Tuple of the hub, port group, port number, and hub port
        """
        hub, port, hub_port = port_info
        if board_revision == BoardRevision.OG:
            if hub:
                return (
                    True,
                    PortGroup.MAIN,
                    REV_OG_USB_PORTS.get(str(port), port),
                    hub_port,
                )
            else:
                return (
                    False,
                    PortGroup.MAIN,
                    REV_OG_USB_PORTS.get(str(port), port),
                    None,
                )
        elif board_revision == BoardRevision.FLEX_B2:
            if hub == FLEX_B2_USB_PORT_GROUP_LEFT:
                port_group = PortGroup.LEFT
                port = FLEX_B2_USB_PORTS.get(str(port), port)
            elif hub == FLEX_B2_USB_PORT_GROUP_RIGHT:
                port_group = PortGroup.RIGHT
                port = FLEX_B2_USB_PORTS.get(str(port), port)
                port = port + 4
            elif hub == FLEX_B2_USB_PORT_GROUP_FRONT:
                port_group = PortGroup.FRONT
            else:
                port_group = PortGroup.UNKNOWN
            if hub_port:
                return True, port_group, port, hub_port
            else:
                return False, port_group, port, None
        else:  # any variant of OT2-Refresh
            if hub_port:
                return True, PortGroup.MAIN, port, hub_port
            else:
                return False, PortGroup.MAIN, port, None

    def __hash__(self) -> int:
        """
        Hash function.

        To have a unique set of nodes, they must
        all have a unique hash. Lists are not
        hashable which is why we need to unpack
        the list here.
        """
        return hash(self.name)


class GPIOPin:
    @classmethod
    def build(cls, name: str, in_out: PinDir, pin: int) -> GPIOPin:
        # use this method if the pin number is the same
        # across all board revisions
        return cls(name, in_out, rev_og=pin, rev_a=pin, rev_b=pin, rev_c=pin)

    @classmethod
    def build_with_rev(
        cls, name: str, in_out: PinDir, **kwargs: Optional[int]
    ) -> GPIOPin:
        return cls(name, in_out, **kwargs)

    def __init__(
        self,
        name: str,
        in_out: PinDir,
        rev_og: Optional[int] = None,
        rev_a: Optional[int] = None,
        rev_b: Optional[int] = None,
        rev_c: Optional[int] = None,
    ):
        self.name = name
        self.in_out = in_out
        self.rev_og = rev_og
        self.rev_a = rev_a
        self.rev_b = rev_b
        self.rev_c = rev_c

    def by_board_rev(self, board_rev: BoardRevision) -> Optional[int]:
        ref = {
            BoardRevision.OG: self.rev_og,
            BoardRevision.A: self.rev_a,
            BoardRevision.B: self.rev_b,
            BoardRevision.C: self.rev_c,
            BoardRevision.UNKNOWN: self.rev_og,
        }
        return ref[board_rev]


class GPIOGroup:
    def __init__(self, pins: List[GPIOPin]):
        self.pins = pins

    def __getattr__(self, item: str) -> GPIOPin:
        res = next(filter(lambda x: x.name is item, self.pins), None)
        assert res, f"Failed to find GPIOPin named: {item}"
        return res

    def by_type(self, pin_dir: PinDir) -> GPIOGroup:
        return GPIOGroup(list(filter(lambda x: x.in_out is pin_dir, self.pins)))

    def by_names(self, names: List[str]) -> GPIOGroup:
        return GPIOGroup(list(filter(lambda x: x.name in names, self.pins)))

    def group_by_pins(self, board_rev: BoardRevision) -> List[List[GPIOPin]]:
        c = groupby(self.pins, key=lambda x: x.by_board_rev(board_rev))
        l: List[List[GPIOPin]] = []
        for k, v in c:
            l.append(list(v))
        return l


gpio_group = GPIOGroup(
    [
        # revision pins (input)
        GPIOPin.build("rev_0", PinDir.rev_input, 17),
        GPIOPin.build("rev_1", PinDir.rev_input, 27),
        # output pins
        GPIOPin.build("frame_leds", PinDir.output, 6),
        GPIOPin.build("blue_button", PinDir.output, 13),
        GPIOPin.build("halt", PinDir.output, 18),
        GPIOPin.build("green_button", PinDir.output, 19),
        GPIOPin.build("audio_enable", PinDir.output, 21),
        GPIOPin.build("isp", PinDir.output, 23),
        GPIOPin.build("reset", PinDir.output, 24),
        GPIOPin.build("red_button", PinDir.output, 26),
        # input pins
        GPIOPin.build("button_input", PinDir.input, 5),
        GPIOPin.build_with_rev("door_sw_filt", PinDir.input, rev_og=20, rev_a=12),
        GPIOPin.build_with_rev("window_sw_filt", PinDir.input, rev_og=20, rev_a=16),
        GPIOPin.build("window_door_sw", PinDir.input, 20),
    ]
)



from __future__ import annotations

from dataclasses import dataclass
from typing_extensions import Protocol as TypingProtocol

from opentrons.types import DeckSlotName
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import requires_version
from opentrons.protocol_engine.clients import SyncClient


# TODO(jbl 2024-02-26) these are hardcoded here since there is a 1 to many relationship going from
#   addressable area names to cutout fixture ids. Currently for trash and waste chute this would not be
#   an issue (trash has only one fixture that provides it, all waste chute fixtures are the same height).
#   The ultimate fix for this is a multiple pass analysis, so for now these are being hardcoded to avoid
#   writing cumbersome guessing logic for area name -> fixture name while still providing a direct link to
#   the numbers in shared data.
_TRASH_BIN_CUTOUT_FIXTURE = "trashBinAdapter"
_TRASH_BIN_OT2_CUTOUT_FIXTURE = "fixedTrashSlot"
_WASTE_CHUTE_CUTOUT_FIXTURE = "wasteChuteRightAdapterCovered"


@dataclass(frozen=True)
class DisposalOffset:
    x: float
    y: float
    z: float


class _DisposalLocation(TypingProtocol):
    """Abstract class for disposal location."""

    def top(self, x: float = 0, y: float = 0, z: float = 0) -> _DisposalLocation:
        """Returns a disposal location with a user set offset."""
        ...

    @property
    def offset(self) -> DisposalOffset:
        """Offset of the disposal location.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        ...

    @property
    def location(self) -> DeckSlotName:
        """Location of the disposal location.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        ...

    @property
    def area_name(self) -> str:
        """Addressable area name of the disposal location.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        ...

    @property
    def height(self) -> float:
        """Height of the disposal location.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        ...


class TrashBin(_DisposalLocation):
    """Represents a Flex or OT-2 trash bin.

    See :py:meth:`.ProtocolContext.load_trash_bin`.
    """

    def __init__(
        self,
        location: DeckSlotName,
        addressable_area_name: str,
        engine_client: SyncClient,
        api_version: APIVersion,
        offset: DisposalOffset = DisposalOffset(x=0, y=0, z=0),
    ) -> None:
        self._location = location
        self._addressable_area_name = addressable_area_name
        self._offset = offset
        self._api_version = api_version
        self._engine_client = engine_client
        if self._engine_client.state.config.robot_type == "OT-2 Standard":
            self._cutout_fixture_name = _TRASH_BIN_OT2_CUTOUT_FIXTURE
        else:
            self._cutout_fixture_name = _TRASH_BIN_CUTOUT_FIXTURE

    @requires_version(2, 18)
    def top(self, x: float = 0, y: float = 0, z: float = 0) -> TrashBin:
        """Add a location offset to a trash bin.

        The default location (``x``, ``y``, and ``z`` all set to ``0``) is the center of
        the bin on the x- and y-axes, and slightly below its physical top on the z-axis.

        Offsets can be positive or negative and are measured in mm.
        See :ref:`protocol-api-deck-coords`.
        """
        return TrashBin(
            location=self._location,
            addressable_area_name=self._addressable_area_name,
            engine_client=self._engine_client,
            api_version=self._api_version,
            offset=DisposalOffset(x=x, y=y, z=z),
        )

    @property
    def offset(self) -> DisposalOffset:
        """Current offset of the trash bin.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        return self._offset

    @property
    def location(self) -> DeckSlotName:
        """Location of the trash bin.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        return self._location

    @property
    def area_name(self) -> str:
        """Addressable area name of the trash bin.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        return self._addressable_area_name

    @property
    def height(self) -> float:
        """Height of the trash bin.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        return self._engine_client.state.addressable_areas.get_fixture_height(
            self._cutout_fixture_name
        )


class WasteChute(_DisposalLocation):
    """Represents a Flex waste chute.

    See :py:meth:`.ProtocolContext.load_waste_chute`.
    """

    def __init__(
        self,
        engine_client: SyncClient,
        api_version: APIVersion,
        offset: DisposalOffset = DisposalOffset(x=0, y=0, z=0),
    ) -> None:
        self._engine_client = engine_client
        self._api_version = api_version
        self._offset = offset

    @requires_version(2, 18)
    def top(self, x: float = 0, y: float = 0, z: float = 0) -> WasteChute:
        """Add a location offset to a waste chute.

        The default location (``x``, ``y``, and ``z`` all set to ``0``) is the center of
        the chute's opening on the x- and y-axes, and slightly below its physical top
        on the z-axis. See :ref:`configure-waste-chute` for more information on possible
        configurations of the chute.

        Offsets can be positive or negative and are measured in mm.
        See :ref:`protocol-api-deck-coords`.
        """
        return WasteChute(
            engine_client=self._engine_client,
            api_version=self._api_version,
            offset=DisposalOffset(x=x, y=y, z=z),
        )

    @property
    def offset(self) -> DisposalOffset:
        """Current offset of the waste chute.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        return self._offset

    @property
    def location(self) -> DeckSlotName:
        """Location of the waste chute.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        return DeckSlotName.SLOT_D3

    @property
    def area_name(self) -> str:
        """Addressable area name of the waste chute.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        # TODO(jbl 2024-02-06) this is hardcoded here and should be removed when a multiple pass analysis exists
        #
        # We want to tell Protocol Engine that there's a waste chute in the waste chute location when it's loaded,
        # so analysis can prevent the user from doing anything that would collide with it. At the same time, we
        # do not want to create a false negative when it comes to addressable area conflict. We therefore use the
        # addressable area `1ChannelWasteChute` because every waste chute cutout fixture provides it and it will
        # provide the engine with the information it needs.
        return "1ChannelWasteChute"

    @property
    def height(self) -> float:
        """Height of the waste chute.

        :meta private:

        This is intended for Opentrons internal use only and is not a guaranteed API.
        """
        return self._engine_client.state.addressable_areas.get_fixture_height(
            _WASTE_CHUTE_CUTOUT_FIXTURE
        )



"""QueueWorker and dependency factory."""
from typing import AsyncGenerator, Callable

from opentrons.hardware_control import HardwareControlAPI
from opentrons.protocol_engine.execution.rail_lights import RailLightsHandler

from ..state.state import StateStore
from ..actions import ActionDispatcher
from ..resources import FileProvider
from .equipment import EquipmentHandler
from .movement import MovementHandler
from .gantry_mover import create_gantry_mover
from .labware_movement import LabwareMovementHandler
from .pipetting import create_pipetting_handler
from .tip_handler import create_tip_handler
from .run_control import RunControlHandler
from .command_executor import CommandExecutor
from .queue_worker import QueueWorker
from .status_bar import StatusBarHandler


def create_queue_worker(
    hardware_api: HardwareControlAPI,
    file_provider: FileProvider,
    state_store: StateStore,
    action_dispatcher: ActionDispatcher,
    command_generator: Callable[[], AsyncGenerator[str, None]],
) -> QueueWorker:
    """Create a ready-to-use QueueWorker instance.

    Arguments:
        hardware_api: Hardware control API to pass down to dependencies.
        file_provider: Provides access to robot server file writing procedures for protocol output.
        state_store: StateStore to pass down to dependencies.
        action_dispatcher: ActionDispatcher to pass down to dependencies.
        error_recovery_policy: ErrorRecoveryPolicy to pass down to dependencies.
        command_generator: Command generator to get the next command to execute.
    """
    gantry_mover = create_gantry_mover(
        hardware_api=hardware_api,
        state_view=state_store,
    )

    equipment_handler = EquipmentHandler(
        hardware_api=hardware_api,
        state_store=state_store,
    )

    movement_handler = MovementHandler(
        hardware_api=hardware_api,
        state_store=state_store,
        gantry_mover=gantry_mover,
    )

    labware_movement_handler = LabwareMovementHandler(
        hardware_api=hardware_api,
        state_store=state_store,
        equipment=equipment_handler,
        movement=movement_handler,
    )

    pipetting_handler = create_pipetting_handler(
        hardware_api=hardware_api,
        state_view=state_store,
    )

    tip_handler = create_tip_handler(
        hardware_api=hardware_api,
        state_view=state_store,
    )

    run_control_handler = RunControlHandler(
        state_store=state_store,
        action_dispatcher=action_dispatcher,
    )
    rail_lights_handler = RailLightsHandler(
        hardware_api=hardware_api,
    )

    status_bar_handler = StatusBarHandler(hardware_api=hardware_api)

    command_executor = CommandExecutor(
        hardware_api=hardware_api,
        file_provider=file_provider,
        state_store=state_store,
        action_dispatcher=action_dispatcher,
        equipment=equipment_handler,
        movement=movement_handler,
        gantry_mover=gantry_mover,
        labware_movement=labware_movement_handler,
        pipetting=pipetting_handler,
        tip_handler=tip_handler,
        run_control=run_control_handler,
        rail_lights=rail_lights_handler,
        status_bar=status_bar_handler,
    )

    return QueueWorker(
        state_store=state_store,
        command_executor=command_executor,
        command_generator=command_generator,
    )



"""A simple pub/sub message broker."""


from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Callable, ContextManager, Generator, Generic, Set, TypeVar


_MessageT = TypeVar("_MessageT")
_CallbackT = Callable[[_MessageT], None]


class ReadOnlyBroker(ABC, Generic[_MessageT]):
    """The read-only subset of `Broker`.

    Useful for typing if you want people to be able to subscribe to your `Broker`,
    but don't want them to be able to publish their own messages to it.
    """

    @abstractmethod
    def subscribed(self, callback: _CallbackT[_MessageT]) -> ContextManager[None]:
        """See `Broker.subscribed()`."""  # noqa: D402
        pass

    @abstractmethod
    def subscribe(self, callback: _CallbackT[_MessageT]) -> Callable[[], None]:
        """See `Broker.subscribe()`."""  # noqa: D402
        pass


class Broker(Generic[_MessageT], ReadOnlyBroker[_MessageT]):
    """A simple pub/sub message broker.

    Subscribers can listen to events. Publishers can push events to all subscribers.
    """

    def __init__(self) -> None:
        self._callbacks: Set[_CallbackT[_MessageT]] = set()

    @contextmanager
    def subscribed(
        self, callback: _CallbackT[_MessageT]
    ) -> Generator[None, None, None]:
        """Register a callback to be called on each message.

        The callback is subscribed when this context manager is entered,
        and unsubscribed when it's exited.

        You must not subscribe the same callback again unless you first usubscribe it.
        """
        unsubscribe = self.subscribe(callback)
        try:
            yield
        finally:
            unsubscribe()

    def subscribe(self, callback: _CallbackT[_MessageT]) -> Callable[[], None]:
        """Register a callback to be called on each message.

        You must not subscribe the same callback again unless you first unsubscribe it.

        Returns:
            A function that you can call to unsubscribe ``callback``.
            You must not call it more than once.
        """

        def unsubscribe() -> None:
            self._callbacks.remove(callback)

        self._callbacks.add(callback)
        return unsubscribe

    def publish(self, message: _MessageT) -> None:
        """Call every subscribed callback, with ``message`` as the argument.

        The order in which the callbacks are called is undefined.

        If any callback raises an exception, it's propagated,
        and any remaining callbacks will be left uncalled.
        """
        # Callback order is undefined because
        # Python sets don't preserve insertion order.
        for callback in self._callbacks:
            callback(message)



# noqa: D100

import typing


def get_union_elements(type_alias: typing.Any) -> tuple[typing.Any, ...]:
    """Return the elements comprising a `typing.Union`.

    e.g. `get_union_elements(Union[int, str, float])` returns `(int, str, float)`.

    If the union itself is wrapped with `typing.Annotated`, it's discarded.
    If an individual element is wrapped with `typing.Annotated`, it's returned as-is.

    Ideally, we would use `typing.get_type_hints(..., include_extras=False)` for this,
    but that only works for getting the annotations of a function, method, module, or
    class, and sometimes we want to unpack an individual union type alias.
    """
    origin, args = typing.get_origin(type_alias), typing.get_args(type_alias)
    if origin is typing.Annotated:
        # Unwrap Annotated[Union[A, B, C]] to Union[A, B, C].
        union_annotation = args[0]
        return get_union_elements(union_annotation)
    elif origin is typing.Union:
        return args
    else:
        raise TypeError(f"get_union_elements() got unrecognized type {type_alias}")



""" opentrons.simulate: functions and entrypoints for simulating protocols

This module has functions that provide a console entrypoint for simulating
a protocol from the command line.
"""
import argparse
import asyncio
import atexit
from contextlib import ExitStack, contextmanager
import sys
import logging
import os
import pathlib
import queue
from typing import (
    TYPE_CHECKING,
    Generator,
    Any,
    Dict,
    List,
    Mapping,
    TextIO,
    Tuple,
    BinaryIO,
    Optional,
    Union,
)
from typing_extensions import Literal

from opentrons_shared_data.robot.types import RobotType

import opentrons
from opentrons import should_use_ot3
from opentrons.hardware_control import (
    API as OT2API,
    ThreadManager,
    ThreadManagedHardware,
)
from opentrons.hardware_control.types import HardwareFeatureFlags

from opentrons.hardware_control.simulator_setup import load_simulator
from opentrons.protocol_api.core.engine import ENGINE_CORE_API_VERSION
from opentrons.protocol_api.protocol_context import ProtocolContext
from opentrons.protocol_engine.create_protocol_engine import (
    create_protocol_engine_in_thread,
    create_protocol_engine,
)
from opentrons.protocol_engine import error_recovery_policy
from opentrons.protocol_engine.state.config import Config
from opentrons.protocol_engine.types import DeckType, EngineStatus, PostRunHardwareState
from opentrons.protocol_reader.protocol_source import ProtocolSource
from opentrons.protocol_runner.protocol_runner import create_protocol_runner, LiveRunner
from opentrons.protocol_runner import RunOrchestrator
from opentrons.protocols.duration import DurationEstimator
from opentrons.protocols.execution import execute
from opentrons.legacy_broker import LegacyBroker
from opentrons.config import IS_ROBOT
from opentrons import protocol_api
from opentrons.legacy_commands import types as command_types

from opentrons.protocols import parse, bundle
from opentrons.protocols.types import (
    ApiDeprecationError,
    Protocol,
    PythonProtocol,
    BundleContents,
)
from opentrons.protocols.api_support.deck_type import (
    for_simulation as deck_type_for_simulation,
    should_load_fixed_trash,
    should_load_fixed_trash_labware_for_python_protocol,
)
from opentrons.protocols.api_support.types import APIVersion
from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from .util import entrypoint_util

if TYPE_CHECKING:
    from opentrons_shared_data.labware.types import (
        LabwareDefinition as LabwareDefinitionDict,
    )


# See Jira RCORE-535.
_JSON_TOO_NEW_MESSAGE = (
    "Protocols created by recent versions of Protocol Designer"
    " cannot currently be simulated with"
    " the opentrons_simulate command-line tool"
    " or the opentrons.simulate.simulate() function."
    " Use the Opentrons App instead."
)


# When a ProtocolContext is using a ProtocolEngine to control the robot,
# it requires some long-lived resources. There's a background thread,
# an asyncio event loop in that thread, and some ProtocolEngine-controlled background tasks in that
# event loop.
#
# When we're executing a protocol file beginning-to-end, we can clean up those resources after it
# completes. However, when someone gets a live ProtocolContext through get_protocol_api(), we have
# no way of knowing when they're done with it. So, as a hack, we keep these resources open
# indefinitely, letting them leak.
#
# We keep this at module scope so that the contained context managers aren't garbage-collected.
# If they're garbage collected, they can close their resources prematurely.
# https://stackoverflow.com/a/69155026/497934
_LIVE_PROTOCOL_ENGINE_CONTEXTS = ExitStack()


# TODO(mm, 2023-10-05): Deduplicate this with opentrons.protocols.parse().
_UserSpecifiedRobotType = Literal["OT-2", "Flex"]
"""The user-facing robot type specifier.

This should match what `opentrons.protocols.parse()` accepts in a protocol's `requirements` dict.
"""


# TODO(mm, 2023-10-05): Type _SimulateResultRunLog more precisely by using TypedDicts from
# opentrons.legacy_commands.
_SimulateResultRunLog = List[Mapping[str, Any]]
_SimulateResult = Tuple[_SimulateResultRunLog, Optional[BundleContents]]


class _AccumulatingHandler(logging.Handler):
    def __init__(
        self,
        level: str,
        command_queue: "queue.Queue[object]",
    ) -> None:
        """Create the handler

        :param level: The logging level to capture
        :param command_queue: The queue.Queue to use for messages
        """
        self._command_queue = command_queue
        super().__init__(level)

    def emit(self, record: object) -> None:
        self._command_queue.put(record)


class _CommandScraper:
    def __init__(
        self, logger: logging.Logger, level: str, broker: LegacyBroker
    ) -> None:
        """An object that handles scraping the broker for commands and integrating log messages
        with them.

        Params:
            logger: The logger to integrate messages from, e.g. ``logging.getLogger("opentrons")``.
            level: The log level to scrape.
            broker: The broker to subscribe to for commands.
        """
        self._logger = logger
        self._level = level
        self._broker = broker
        self._commands: _SimulateResultRunLog = []

    @property
    def commands(self) -> _SimulateResultRunLog:
        """The list of commands scraped while `.scrape()` was open, integrated with log messages.

        See :py:obj:`simulate` for the return type.
        """
        return self._commands

    @contextmanager
    def scrape(self) -> Generator[None, None, None]:
        """While this context manager is open, scrape the broker for commands and integrate log
        messages with them. The accumulated commands will be accessible through `.commands`.
        """
        log_queue: "queue.Queue[object]" = queue.Queue()

        depth = 0

        def handle_command(message: command_types.CommandMessage) -> None:
            """The callback that we will subscribe to the broker."""
            nonlocal depth
            payload = message["payload"]
            if message["$"] == "before":
                self._commands.append({"level": depth, "payload": payload, "logs": []})
                depth += 1
            else:
                while not log_queue.empty():
                    self._commands[-1]["logs"].append(log_queue.get())
                depth = max(depth - 1, 0)

        if self._level != "none":
            # The simulation entry points probably leave logging unconfigured, so the level will be
            # Python's default. Set it to what the user asked to make sure we see the expected
            # records.
            #
            # TODO(mm, 2023-10-03): This is a bit too intrusive for something whose job is just to
            # "scrape." The entry point function should be responsible for setting the underlying
            # logger's level. Also, we should probably restore the original level when we're done.
            level = getattr(logging, self._level.upper(), logging.WARNING)
            self._logger.setLevel(level)

            log_handler: Optional[_AccumulatingHandler] = _AccumulatingHandler(
                self._level.upper(), log_queue
            )
        else:
            log_handler = None

        with ExitStack() as exit_stack:
            if log_handler is not None:
                self._logger.addHandler(log_handler)
                exit_stack.callback(self._logger.removeHandler, log_handler)

            unsubscribe_from_broker = self._broker.subscribe(
                command_types.COMMAND, handle_command
            )
            exit_stack.callback(unsubscribe_from_broker)

            yield


def get_protocol_api(
    version: Union[str, APIVersion],
    bundled_labware: Optional[Dict[str, "LabwareDefinitionDict"]] = None,
    bundled_data: Optional[Dict[str, bytes]] = None,
    extra_labware: Optional[Dict[str, "LabwareDefinitionDict"]] = None,
    hardware_simulator: Optional[ThreadManagedHardware] = None,
    # Additional arguments are kw-only to make mistakes harder in environments without
    # type checking, like Jupyter Notebook.
    *,
    robot_type: Optional[_UserSpecifiedRobotType] = None,
    use_virtual_hardware: bool = True,
) -> protocol_api.ProtocolContext:
    """
    Build and return a ``protocol_api.ProtocolContext`` that simulates robot control.

    This can be used to simulate protocols from interactive Python sessions
    such as Jupyter or an interpreter on the command line:

    .. code-block:: python

        >>> from opentrons.simulate import get_protocol_api
        >>> protocol = get_protocol_api('2.0')
        >>> instr = protocol.load_instrument('p300_single', 'right')
        >>> instr.home()

    :param version: The API version to use. This must be lower than
        ``opentrons.protocol_api.MAX_SUPPORTED_VERSION``.
        It may be specified either as a string (``'2.0'``) or
        as a ``protocols.types.APIVersion``
        (``APIVersion(2, 0)``).
    :param bundled_labware: If specified, a mapping from labware names to
        labware definitions for labware to consider in the
        protocol. Note that if you specify this, *only*
        labware in this argument will be allowed in the
        protocol. This is preparation for a beta feature
        and is best not used.
    :param bundled_data: If specified, a mapping from filenames to contents
        for data to be available in the protocol from
        :py:obj:`opentrons.protocol_api.ProtocolContext.bundled_data`.
    :param extra_labware: A mapping from labware load names to custom labware definitions.
        If this is ``None`` (the default), and this function is called on a robot,
        it will look for labware in the ``labware`` subdirectory of the Jupyter
        data directory.
    :param hardware_simulator: This is only for internal use by Opentrons. If specified,
        it's a hardware simulator instance to reuse instead of creating a fresh one.
    :param robot_type: The type of robot to simulate: either ``"Flex"`` or ``"OT-2"``.
        If you're running this function on a robot, the default is the type of that
        robot. Otherwise, the default is ``"OT-2"``, for backwards compatibility.
    :param use_virtual_hardware: This is only for internal use by Opentrons.
        If ``True``, use the Protocol Engine's virtual hardware. If ``False``, use the
        lower level hardware simulator.
    :return: The protocol context.
    """
    if isinstance(version, str):
        checked_version = parse.version_from_string(version)
    elif not isinstance(version, APIVersion):
        raise TypeError("version must be either a string or an APIVersion")
    else:
        checked_version = version

    current_robot_type = _get_current_robot_type()
    if robot_type is None:
        if current_robot_type is None:
            parsed_robot_type: RobotType = "OT-2 Standard"
        else:
            parsed_robot_type = current_robot_type
    else:
        # TODO(mm, 2023-10-09): This raises a slightly wrong error message, mentioning the camelCase
        # `robotType` field in Python files instead of the snake_case `robot_type` argument for this
        # function.
        parsed_robot_type = parse.robot_type_from_python_identifier(robot_type)
    _validate_can_simulate_for_robot_type(parsed_robot_type)
    deck_type = deck_type_for_simulation(parsed_robot_type)

    if extra_labware is None:
        extra_labware = {
            uri: details.definition
            for uri, details in (entrypoint_util.find_jupyter_labware() or {}).items()
        }

    checked_hardware = _make_hardware_simulator(
        override=hardware_simulator, robot_type=parsed_robot_type
    )

    if checked_version < ENGINE_CORE_API_VERSION:
        context = _create_live_context_non_pe(
            api_version=checked_version,
            deck_type=deck_type,
            hardware_api=checked_hardware,
            bundled_labware=bundled_labware,
            bundled_data=bundled_data,
            extra_labware=extra_labware,
        )
    else:
        if bundled_labware is not None:
            # Protocol Engine has a deep assumption that standard labware definitions are always
            # implicitly loadable.
            raise NotImplementedError(
                f"The bundled_labware argument is not currently supported for Python protocols"
                f" with apiLevel {ENGINE_CORE_API_VERSION} or newer."
            )
        context = _create_live_context_pe(
            api_version=checked_version,
            robot_type=parsed_robot_type,
            deck_type=deck_type,
            hardware_api=checked_hardware,
            bundled_data=bundled_data,
            extra_labware=extra_labware,
            use_pe_virtual_hardware=use_virtual_hardware,
        )

    # Intentional difference from execute.get_protocol_api():
    # For the caller's convenience, we home the virtual hardware so they don't get MustHomeErrors.
    # Since this hardware is virtual, there's no harm in commanding this "movement" implicitly.
    #
    # Calling `checked_hardware_sync.home()` is a hack. It ought to be redundant with
    # `context.home()`. We need it here to work around a Protocol Engine simulation bug
    # where both the `HardwareControlAPI` level and the `ProtocolEngine` level need to
    # be homed for certain commands to work. https://opentrons.atlassian.net/browse/EXEC-646
    checked_hardware.sync.home()
    context.home()

    return context


def _make_hardware_simulator(
    override: Optional[ThreadManagedHardware], robot_type: RobotType
) -> ThreadManagedHardware:
    if override:
        return override
    elif robot_type == "OT-3 Standard":
        # Local import because this isn't available on OT-2s.
        from opentrons.hardware_control.ot3api import OT3API

        return ThreadManager(
            OT3API.build_hardware_simulator,
            feature_flags=HardwareFeatureFlags.build_from_ff(),
        )
    elif robot_type == "OT-2 Standard":
        return ThreadManager(
            OT2API.build_hardware_simulator,
            feature_flags=HardwareFeatureFlags.build_from_ff(),
        )


@contextmanager
def _make_hardware_simulator_cm(
    config_file_path: Optional[pathlib.Path], robot_type: RobotType
) -> Generator[ThreadManagedHardware, None, None]:
    if config_file_path is not None:
        result = ThreadManager(
            load_simulator,
            pathlib.Path(config_file_path),
        )
        try:
            yield result
        finally:
            result.clean_up()
    else:
        result = _make_hardware_simulator(override=None, robot_type=robot_type)
        try:
            yield result
        finally:
            result.clean_up()


def _get_current_robot_type() -> Optional[RobotType]:
    """Return the type of robot that we're running on, or None if we're not on a robot."""
    if IS_ROBOT:
        return "OT-3 Standard" if should_use_ot3() else "OT-2 Standard"
    else:
        return None


def _validate_can_simulate_for_robot_type(robot_type: RobotType) -> None:
    """Raise if this device cannot simulate protocols written for the given robot type."""
    current_robot_type = _get_current_robot_type()
    if current_robot_type is None:
        # When installed locally, this package can simulate protocols for any robot type.
        pass
    elif robot_type != current_robot_type:
        # Match robot server behavior: raise an early error if we're on a robot and the caller
        # tries to simulate a protocol written for a different robot type.

        # FIXME: This exposes the internal strings "OT-2 Standard" and "OT-3 Standard".
        # https://opentrons.atlassian.net/browse/RSS-370
        raise RuntimeError(
            f'This robot is of type "{current_robot_type}",'
            f' so it can\'t simulate protocols for robot type "{robot_type}"'
        )


def bundle_from_sim(
    protocol: PythonProtocol, context: opentrons.protocol_api.ProtocolContext
) -> BundleContents:
    """
    From a protocol, and the context that has finished simulating that
    protocol, determine what needs to go in a bundle for the protocol.
    """
    bundled_labware: Dict[str, "LabwareDefinitionDict"] = {}
    for lw in context.loaded_labwares.values():
        if (
            isinstance(lw, opentrons.protocol_api.labware.Labware)
            and lw.uri not in bundled_labware
        ):
            bundled_labware[lw.uri] = lw._core.get_definition()

    return BundleContents(
        protocol.text,
        bundled_data=context.bundled_data,
        bundled_labware=bundled_labware,
        bundled_python={},
    )


def simulate(
    protocol_file: Union[BinaryIO, TextIO],
    file_name: Optional[str] = None,
    custom_labware_paths: Optional[List[str]] = None,
    custom_data_paths: Optional[List[str]] = None,
    propagate_logs: bool = False,
    hardware_simulator_file_path: Optional[str] = None,
    duration_estimator: Optional[DurationEstimator] = None,
    log_level: str = "warning",
) -> _SimulateResult:
    """
    Simulate the protocol itself.

    This is a one-stop function to simulate a protocol, whether Python or JSON,
    no matter the API version, from external (i.e. not bound up in other
    internal server infrastructure) sources.

    To simulate an opentrons protocol from other places, pass in a file-like
    object as ``protocol_file``; this function either returns (if the simulation
    has no problems) or raises an exception.

    To call from the command line, use either the autogenerated entrypoint
    ``opentrons_simulate`` (``opentrons_simulate.exe``, on windows) or
    ``python -m opentrons.simulate``.

    The return value is the run log, a list of dicts that represent the
    commands executed by the robot; and either the contents of the protocol
    that would be required to bundle, or ``None``.

    Each dict element in the run log has the following keys:

        - ``level``: The depth at which this command is nested. If this an
          aspirate inside a mix inside a transfer, for instance, it would be 3.

        - ``payload``: The command. The human-readable run log text is available at
          ``payload["text"]``. The other keys of ``payload`` are command-dependent;
          see ``opentrons.legacy_commands``.

          .. note::
            In older software versions, ``payload["text"]`` was a
            `format string <https://docs.python.org/3/library/string.html#formatstrings>`_.
            To get human-readable text, you had to do ``payload["text"].format(**payload)``.
            Don't do that anymore. If ``payload["text"]`` happens to contain any
            ``{`` or ``}`` characters, it can confuse ``.format()`` and cause it to raise a
            ``KeyError``.

        - ``logs``: Any log messages that occurred during execution of this
          command, as a standard Python :py:obj:`~logging.LogRecord`.

    :param protocol_file: The protocol file to simulate.
    :param file_name: The name of the file
    :param custom_labware_paths: A list of directories to search for custom labware.
        Loads valid labware from these paths and makes them available
        to the protocol context. If this is ``None`` (the default), and
        this function is called on a robot, it will look in the ``labware``
        subdirectory of the Jupyter data directory.
    :param custom_data_paths: A list of directories or files to load custom
        data files from. Ignored if the apiv2 feature
        flag if not set. Entries may be either files or
        directories. Specified files and the
        non-recursive contents of specified directories
        are presented by the protocol context in
        ``protocol_api.ProtocolContext.bundled_data``.
    :param hardware_simulator_file_path: A path to a JSON file defining the simulated
        hardware. This is mainly for internal use by Opentrons, and is not necessary
        to simulate protocols.
    :param duration_estimator: For internal use only.
        Optional duration estimator object.
    :param propagate_logs: Whether this function should allow logs from the
        Opentrons stack to propagate up to the root handler.
        This can be useful if you're integrating this
        function in a larger application, but most logs that
        occur during protocol simulation are best associated
        with the actions in the protocol that cause them.
        Default: ``False``
    :param log_level: The level of logs to capture in the run log:
        ``"debug"``, ``"info"``, ``"warning"``, or ``"error"``.
        Defaults to ``"warning"``.
    :returns: A tuple of a run log for user output, and possibly the required
        data to write to a bundle to bundle this protocol. The bundle is
        only emitted if bundling is allowed
        and this is an unbundled Protocol API
        v2 python protocol. In other cases it is None.
    """
    stack_logger = logging.getLogger("opentrons")
    stack_logger.propagate = propagate_logs
    # _CommandScraper will set the level of this logger.

    # TODO(mm, 2023-10-02): Switch this truthy check to `is not None`
    # to match documented behavior.
    # See notes in https://github.com/Opentrons/opentrons/pull/13107
    if custom_labware_paths:
        extra_labware = entrypoint_util.labware_from_paths(custom_labware_paths)
    else:
        extra_labware = entrypoint_util.find_jupyter_labware() or {}

    if custom_data_paths:
        extra_data = entrypoint_util.datafiles_from_paths(custom_data_paths)
    else:
        extra_data = {}

    contents = protocol_file.read()
    try:
        protocol = parse.parse(
            contents,
            file_name,
            extra_labware={
                uri: details.definition for uri, details in extra_labware.items()
            },
            extra_data=extra_data,
        )
    except parse.JSONSchemaVersionTooNewError as e:
        # opentrons.protocols.parse() doesn't support new JSON protocols.
        # The code to do that should be moved from opentrons.protocol_reader.
        # See https://opentrons.atlassian.net/browse/PLAT-94.
        raise NotImplementedError(_JSON_TOO_NEW_MESSAGE) from e

    if protocol.api_level < APIVersion(2, 0):
        raise ApiDeprecationError(version=protocol.api_level)

    _validate_can_simulate_for_robot_type(protocol.robot_type)

    with _make_hardware_simulator_cm(
        config_file_path=(
            None
            if hardware_simulator_file_path is None
            else pathlib.Path(hardware_simulator_file_path)
        ),
        robot_type=protocol.robot_type,
    ) as hardware_simulator:
        if protocol.api_level < ENGINE_CORE_API_VERSION:
            return _run_file_non_pe(
                protocol=protocol,
                hardware_api=hardware_simulator,
                logger=stack_logger,
                level=log_level,
                duration_estimator=duration_estimator,
            )
        else:
            # TODO(mm, 2023-07-06): Once these NotImplementedErrors are resolved, consider removing
            # the enclosing if-else block and running everything through _run_file_pe() for simplicity.
            if custom_data_paths:
                raise NotImplementedError(
                    f"The custom_data_paths argument is not currently supported for Python protocols"
                    f" with apiLevel {ENGINE_CORE_API_VERSION} or newer."
                )
            protocol_file.seek(0)
            return _run_file_pe(
                protocol=protocol,
                robot_type=protocol.robot_type,
                hardware_api=hardware_simulator,
                stack_logger=stack_logger,
                log_level=log_level,
            )


def format_runlog(runlog: List[Mapping[str, Any]]) -> str:
    """
    Format a run log (return value of :py:obj:`simulate`) into a
    human-readable string

    :param runlog: The output of a call to :py:obj:`simulate`
    """
    to_ret = []
    for command in runlog:
        to_ret.append("\t" * command["level"] + command["payload"].get("text", ""))
        if command["logs"]:
            to_ret.append("\t" * command["level"] + "Logs from this command:")
            to_ret.extend(
                [
                    "\t" * command["level"]
                    + f"{l.levelname} ({l.module}): {l.msg}" % l.args
                    for l in command["logs"]  # noqa: E741
                ]
            )
    return "\n".join(to_ret)


def _get_bundle_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser.add_argument(
        "-b",
        "--bundle",
        nargs="?",
        const="PROTOCOL.ot2.zip",
        default=None,
        action="store",
        type=str,
        help="Bundle the specified protocol file, any labware used in it, and "
        "any files in the data directories specified with -D into a "
        "bundle. This bundle can be executed on a robot and carries with "
        "it all the custom labware and data required to run. Without a "
        "value specified in this argument, the bundle will be called "
        "(protocol name without the .py).ot2.zip, but you can specify "
        "a different output name. \n"
        "These bundles are a beta feature, and their behavior may change",
    )
    return parser


def allow_bundle() -> bool:
    """
    Check if bundling is allowed with a special not-exposed-to-the-app flag.

    Returns ``True`` if the environment variable
    ``OT_API_FF_allowBundleCreation`` is ``"1"``
    """
    return os.getenv("OT_API_FF_allowBundleCreation") == "1"


def get_arguments(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Get the argument parser for this module

    Useful if you want to use this module as a component of another CLI program
    and want to add its arguments.

    :param parser: A parser to add arguments to. If not specified, one will be created.
    :returns argparse.ArgumentParser: The parser with arguments added.
    """
    parser.add_argument(
        "-l",
        "--log-level",
        choices=["debug", "info", "warning", "error", "none"],
        default="warning",
        help="Specify the level filter for logs to show on the command line. "
        'Log levels below warning can be chatty. If "none", do not show logs',
    )

    parser.add_argument(
        "-L",
        "--custom-labware-path",
        action="append",
        default=[os.getcwd()],
        help="Specify directories to search for custom labware definitions. "
        "You can specify this argument multiple times. Once you specify "
        "a directory in this way, labware definitions in that directory "
        "will become available in ProtocolContext.load_labware(). "
        "Only directories specified directly by "
        "this argument are searched, not their children. JSON files that "
        "do not define labware will be ignored with a message. "
        "The current directory (the one from which you are "
        "invoking this program) will always be included implicitly, "
        "in addition to any directories that you specify.",
    )
    parser.add_argument(
        "-D",
        "--custom-data-path",
        action="append",
        nargs="?",
        const=".",
        default=[],
        help="Specify directories to search for custom data files. "
        "You can specify this argument multiple times. Once you specify "
        "a directory in this way, files located in the specified "
        "directory will be available in ProtocolContext.bundled_data. "
        "Note that bundle execution will still only allow data files in "
        "the bundle. If you specify this without a path, it will "
        "add the current path implicitly. If you do not specify this "
        "argument at all, no data files will be added. Any file in the "
        "specified paths will be loaded into memory and included in the "
        "bundle if --bundle is passed, so be careful that any directory "
        "you specify has only the files you want. It is usually a "
        "better idea to use -d so no files are accidentally included. "
        "Also note that data files are made available as their name, not "
        "their full path, so name them uniquely.",
    )
    parser.add_argument(
        "-s",
        "--custom-hardware-simulator-file",
        type=str,
        default=None,
        help="Specify a file that describes the features present in the "
        "hardware simulator. Features can be instruments, modules, and "
        "configuration.",
    )
    parser.add_argument(
        "-d",
        "--custom-data-file",
        action="append",
        default=[],
        help="Specify data files to be made available in "
        "ProtocolContext.bundled_data (and possibly bundled if --bundle "
        "is passed). Can be specified multiple times with different "
        "files. It is usually a better idea to use this than -D because "
        "there is less possibility of accidentally including something.",
    )
    if allow_bundle():
        parser = _get_bundle_args(parser)

    parser.add_argument(
        "-e",
        "--estimate-duration",
        action="store_true",
        # TODO (AL, 2021-07-26): Better wording.
        help="Estimate how long the protocol will take to complete."
        " This is an experimental feature.",
    )

    parser.add_argument(
        "protocol",
        metavar="PROTOCOL",
        type=argparse.FileType("rb"),
        help="The protocol file to simulate. If you pass '-', you can pipe "
        "the protocol via stdin; this could be useful if you want to use this "
        "utility as part of an automated workflow.",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {opentrons.__version__}",
        help="Print the opentrons package version and exit",
    )
    parser.add_argument(
        "-o",
        "--output",
        action="store",
        help="What to output during simulations",
        choices=["runlog", "nothing"],
        default="runlog",
    )
    return parser


def _get_bundle_dest(
    bundle_name: Optional[str], default_key: str, proto_name: str
) -> Optional[BinaryIO]:
    if bundle_name == default_key:
        protopath = pathlib.Path(proto_name)
        # strip all the suffixes since protocols are often named
        # .ot2.zip
        if protopath.name.endswith(".ot2.py"):
            protoname = pathlib.Path(protopath.stem).stem
        else:
            protoname = protopath.stem
        bundle_name = str((pathlib.Path.cwd() / protoname).with_suffix(".ot2.zip"))
        return open(bundle_name, "wb")
    elif bundle_name:
        return open(bundle_name, "wb")
    else:
        return None


def _create_live_context_non_pe(
    api_version: APIVersion,
    hardware_api: ThreadManagedHardware,
    deck_type: str,
    extra_labware: Optional[Dict[str, "LabwareDefinitionDict"]],
    bundled_labware: Optional[Dict[str, "LabwareDefinitionDict"]],
    bundled_data: Optional[Dict[str, bytes]],
) -> ProtocolContext:
    """Return a live ProtocolContext.

    This controls the robot through the older infrastructure, instead of through Protocol Engine.
    """
    assert api_version < ENGINE_CORE_API_VERSION
    return protocol_api.create_protocol_context(
        api_version=api_version,
        deck_type=deck_type,
        hardware_api=hardware_api,
        bundled_labware=bundled_labware,
        bundled_data=bundled_data,
        extra_labware=extra_labware,
    )


def _create_live_context_pe(
    api_version: APIVersion,
    hardware_api: ThreadManagedHardware,
    robot_type: RobotType,
    deck_type: str,
    extra_labware: Dict[str, "LabwareDefinitionDict"],
    bundled_data: Optional[Dict[str, bytes]],
    use_pe_virtual_hardware: bool = True,
) -> ProtocolContext:
    """Return a live ProtocolContext that controls the robot through ProtocolEngine."""
    assert api_version >= ENGINE_CORE_API_VERSION
    hardware_api_wrapped = hardware_api.wrapped()
    global _LIVE_PROTOCOL_ENGINE_CONTEXTS
    pe, loop = _LIVE_PROTOCOL_ENGINE_CONTEXTS.enter_context(
        create_protocol_engine_in_thread(
            hardware_api=hardware_api_wrapped,
            config=_get_protocol_engine_config(
                robot_type, use_pe_virtual_hardware=use_pe_virtual_hardware
            ),
            deck_configuration=None,
            file_provider=None,
            error_recovery_policy=error_recovery_policy.never_recover,
            drop_tips_after_run=False,
            post_run_hardware_state=PostRunHardwareState.STAY_ENGAGED_IN_PLACE,
            load_fixed_trash=should_load_fixed_trash_labware_for_python_protocol(
                api_version
            ),
        )
    )

    # `async def` so we can use loop.run_coroutine_threadsafe() to wait for its completion.
    # Non-async would use call_soon_threadsafe(), which makes the waiting harder.
    async def add_all_extra_labware() -> None:
        for labware_definition_dict in extra_labware.values():
            labware_definition = LabwareDefinition.parse_obj(labware_definition_dict)
            pe.add_labware_definition(labware_definition)

    # Add extra_labware to ProtocolEngine, being careful not to modify ProtocolEngine from this
    # thread. See concurrency notes in ProtocolEngine docstring.
    future = asyncio.run_coroutine_threadsafe(add_all_extra_labware(), loop)
    future.result()

    return protocol_api.create_protocol_context(
        api_version=api_version,
        hardware_api=hardware_api,
        deck_type=deck_type,
        protocol_engine=pe,
        protocol_engine_loop=loop,
        bundled_data=bundled_data,
    )


def _run_file_non_pe(
    protocol: Protocol,
    hardware_api: ThreadManagedHardware,
    logger: logging.Logger,
    level: str,
    duration_estimator: Optional[DurationEstimator],
) -> _SimulateResult:
    """Run a protocol file without Protocol Engine, with the older infrastructure instead."""
    if isinstance(protocol, PythonProtocol):
        extra_labware = protocol.extra_labware
        bundled_labware = protocol.bundled_labware
        bundled_data = protocol.bundled_data
    else:
        # JSON protocols do have "bundled labware" embedded in them, but those aren't represented in
        # the parsed Protocol object and we don't need to create the ProtocolContext with them.
        # execute_apiv2.run_protocol() will pull them out of the JSON and load them into the
        # ProtocolContext.
        extra_labware = None
        bundled_labware = None
        bundled_data = None

    context = _create_live_context_non_pe(
        api_version=protocol.api_level,
        hardware_api=hardware_api,
        deck_type=deck_type_for_simulation(protocol.robot_type),
        extra_labware=extra_labware,
        bundled_labware=bundled_labware,
        bundled_data=bundled_data,
    )

    scraper = _CommandScraper(logger=logger, level=level, broker=context.broker)
    if duration_estimator:
        context.broker.subscribe(command_types.COMMAND, duration_estimator.on_message)

    context.home()
    with scraper.scrape():
        try:
            execute.run_protocol(
                protocol, context, run_time_parameters_with_overrides=None
            )
            if (
                isinstance(protocol, PythonProtocol)
                and protocol.api_level >= APIVersion(2, 0)
                and protocol.bundled_labware is None
                and allow_bundle()
            ):
                bundle_contents: Optional[BundleContents] = bundle_from_sim(
                    protocol, context
                )
            else:
                bundle_contents = None

        finally:
            context.cleanup()

    return scraper.commands, bundle_contents


def _run_file_pe(
    protocol: Protocol,
    robot_type: RobotType,
    hardware_api: ThreadManagedHardware,
    stack_logger: logging.Logger,
    log_level: str,
) -> _SimulateResult:
    """Run a protocol file with Protocol Engine."""
    # TODO (spp, 2024-03-18): use run-time param overrides once enabled for cli protocol simulation.

    async def run(protocol_source: ProtocolSource) -> _SimulateResult:
        hardware_api_wrapped = hardware_api.wrapped()
        protocol_engine = await create_protocol_engine(
            hardware_api=hardware_api_wrapped,
            config=_get_protocol_engine_config(
                robot_type, use_pe_virtual_hardware=True
            ),
            error_recovery_policy=error_recovery_policy.never_recover,
            load_fixed_trash=should_load_fixed_trash(protocol_source.config),
        )

        protocol_runner = create_protocol_runner(
            protocol_config=protocol_source.config,
            protocol_engine=protocol_engine,
            hardware_api=hardware_api_wrapped,
        )

        orchestrator = RunOrchestrator(
            hardware_api=hardware_api_wrapped,
            protocol_engine=protocol_engine,
            json_or_python_protocol_runner=protocol_runner,
            fixit_runner=LiveRunner(
                protocol_engine=protocol_engine, hardware_api=hardware_api_wrapped
            ),
            setup_runner=LiveRunner(
                protocol_engine=protocol_engine, hardware_api=hardware_api_wrapped
            ),
            protocol_live_runner=LiveRunner(
                protocol_engine=protocol_engine, hardware_api=hardware_api_wrapped
            ),
        )

        # TODO(mm, 2024-08-06): This home is theoretically redundant with Protocol
        # Engine `home` commands within the `RunOrchestrator`. However, we need this to
        # work around Protocol Engine bugs where both the `HardwareControlAPI` level
        # and the `ProtocolEngine` level need to be homed for certain commands to work.
        # https://opentrons.atlassian.net/browse/EXEC-646
        await hardware_api_wrapped.home()

        scraper = _CommandScraper(stack_logger, log_level, protocol_runner.broker)
        with scraper.scrape():
            result = await orchestrator.run(
                # deck_configuration=[] is a placeholder value, ignored because
                # the Protocol Engine config specifies use_simulated_deck_config=True.
                deck_configuration=[],
                protocol_source=protocol_source,
            )

        if result.state_summary.status != EngineStatus.SUCCEEDED:
            raise entrypoint_util.ProtocolEngineExecuteError(
                result.state_summary.errors
            )

        # We don't currently support returning bundle contents from protocols run through
        # Protocol Engine. To get them, bundle_from_sim() requires direct access to the
        # ProtocolContext, which opentrons.protocol_runner does not grant us.
        bundle_contents = None

        return scraper.commands, bundle_contents

    with entrypoint_util.adapt_protocol_source(protocol) as protocol_source:
        return asyncio.run(run(protocol_source))


def _get_protocol_engine_config(
    robot_type: RobotType, use_pe_virtual_hardware: bool
) -> Config:
    """Return a Protocol Engine config to execute protocols on this device."""
    return Config(
        robot_type=robot_type,
        deck_type=DeckType(deck_type_for_simulation(robot_type)),
        ignore_pause=True,
        use_virtual_pipettes=use_pe_virtual_hardware,
        use_virtual_modules=use_pe_virtual_hardware,
        use_virtual_gripper=use_pe_virtual_hardware,
        use_simulated_deck_config=True,
    )


@atexit.register
def _clear_live_protocol_engine_contexts() -> None:
    global _LIVE_PROTOCOL_ENGINE_CONTEXTS
    _LIVE_PROTOCOL_ENGINE_CONTEXTS.close()


# Note - this script is also set up as a setuptools entrypoint and thus does
# an absolute minimum of work since setuptools does something odd generating
# the scripts
def main() -> int:
    """Run the simulation"""
    parser = argparse.ArgumentParser(
        prog="opentrons_simulate",
        description="Simulate a protocol for an Opentrons robot",
    )
    parser = get_arguments(parser)

    args = parser.parse_args()

    # TODO(mm, 2022-12-01): Configure the DurationEstimator with the correct deck type.
    duration_estimator = DurationEstimator() if args.estimate_duration else None

    try:
        runlog, maybe_bundle = simulate(
            protocol_file=args.protocol,
            file_name=args.protocol.name,
            custom_labware_paths=args.custom_labware_path,
            custom_data_paths=(args.custom_data_path + args.custom_data_file),
            duration_estimator=duration_estimator,
            hardware_simulator_file_path=getattr(
                args, "custom_hardware_simulator_file"
            ),
            log_level=args.log_level,
        )
    except entrypoint_util.ProtocolEngineExecuteError as error:
        print(error.to_stderr_string(), file=sys.stderr)
        return 1

    if maybe_bundle:
        bundle_name = getattr(args, "bundle", None)
        if bundle_name == args.protocol.name:
            raise RuntimeError("Bundle path and input path must be different")
        bundle_dest = _get_bundle_dest(
            bundle_name, "PROTOCOL.ot2.zip", args.protocol.name
        )
        if bundle_dest:
            bundle.create_bundle(maybe_bundle, bundle_dest)

    if args.output == "runlog":
        print(format_runlog(runlog))

    if duration_estimator:
        duration_seconds = duration_estimator.get_total_duration()
        hours = int(duration_seconds / 60 / 60)
        minutes = int((duration_seconds % (60 * 60)) / 60)
        print("--------------------------------------------------------------")
        print(f"Estimated protocol duration: {hours}h:{minutes}m")
        print("--------------------------------------------------------------")
        print("WARNING: Protocol duration estimation is an experimental feature")

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Basic pipette data state and store."""
from __future__ import annotations

import dataclasses
from typing import (
    Dict,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from opentrons_shared_data.pipette import pipette_definition
from opentrons.config.defaults_ot2 import Z_RETRACT_DISTANCE
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.hardware_control import CriticalPoint
from opentrons.hardware_control.nozzle_manager import (
    NozzleConfigurationType,
    NozzleMap,
)
from opentrons.types import MountType, Mount as HwMount, Point

from . import update_types
from .. import commands
from .. import errors
from ..types import (
    LoadedPipette,
    MotorAxis,
    FlowRates,
    DeckPoint,
    CurrentWell,
    CurrentAddressableArea,
    CurrentPipetteLocation,
    TipGeometry,
)
from ..actions import (
    Action,
    FailCommandAction,
    SetPipetteMovementSpeedAction,
    SucceedCommandAction,
    get_state_updates,
)
from ._abstract_store import HasState, HandlesActions


@dataclasses.dataclass(frozen=True)
class HardwarePipette:
    """Hardware pipette data."""

    mount: HwMount
    config: PipetteDict


@dataclasses.dataclass(frozen=True)
class CurrentDeckPoint:
    """The latest deck point and mount the robot has accessed."""

    mount: Optional[MountType]
    deck_point: Optional[DeckPoint]


@dataclasses.dataclass(frozen=True)
class BoundingNozzlesOffsets:
    """Offsets of the bounding nozzles of the pipette."""

    back_left_offset: Point
    front_right_offset: Point


@dataclasses.dataclass(frozen=True)
class PipetteBoundingBoxOffsets:
    """Offsets of the corners of the pipette's bounding box."""

    back_left_corner: Point
    front_right_corner: Point
    back_right_corner: Point
    front_left_corner: Point


@dataclasses.dataclass(frozen=True)
class StaticPipetteConfig:
    """Static config for a pipette."""

    model: str
    serial_number: str
    display_name: str
    min_volume: float
    max_volume: float
    channels: int
    tip_configuration_lookup_table: Dict[
        float, pipette_definition.SupportedTipsDefinition
    ]
    nominal_tip_overlap: Dict[str, float]
    home_position: float
    nozzle_offset_z: float
    pipette_bounding_box_offsets: PipetteBoundingBoxOffsets
    bounding_nozzle_offsets: BoundingNozzlesOffsets
    default_nozzle_map: NozzleMap  # todo(mm, 2024-10-14): unused, remove?
    lld_settings: Optional[Dict[str, Dict[str, float]]]


@dataclasses.dataclass
class PipetteState:
    """Basic pipette data state and getter methods."""

    # todo(mm, 2024-10-14): It's getting difficult to ensure that all of these
    # attributes are populated at the appropriate times. Refactor to a
    # single dict-of-many-things instead of many dicts-of-single-things.
    pipettes_by_id: Dict[str, LoadedPipette]
    aspirated_volume_by_id: Dict[str, Optional[float]]
    current_location: Optional[CurrentPipetteLocation]
    current_deck_point: CurrentDeckPoint
    attached_tip_by_id: Dict[str, Optional[TipGeometry]]
    movement_speed_by_id: Dict[str, Optional[float]]
    static_config_by_id: Dict[str, StaticPipetteConfig]
    flow_rates_by_id: Dict[str, FlowRates]
    nozzle_configuration_by_id: Dict[str, NozzleMap]
    liquid_presence_detection_by_id: Dict[str, bool]


class PipetteStore(HasState[PipetteState], HandlesActions):
    """Pipette state container."""

    _state: PipetteState

    def __init__(self) -> None:
        """Initialize a PipetteStore and its state."""
        self._state = PipetteState(
            pipettes_by_id={},
            aspirated_volume_by_id={},
            attached_tip_by_id={},
            current_location=None,
            current_deck_point=CurrentDeckPoint(mount=None, deck_point=None),
            movement_speed_by_id={},
            static_config_by_id={},
            flow_rates_by_id={},
            nozzle_configuration_by_id={},
            liquid_presence_detection_by_id={},
        )

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        for state_update in get_state_updates(action):
            self._set_load_pipette(state_update)
            self._update_current_location(state_update)
            self._update_pipette_config(state_update)
            self._update_pipette_nozzle_map(state_update)
            self._update_tip_state(state_update)

        if isinstance(action, (SucceedCommandAction, FailCommandAction)):
            self._update_volumes(action)

        elif isinstance(action, SetPipetteMovementSpeedAction):
            self._state.movement_speed_by_id[action.pipette_id] = action.speed

    def _set_load_pipette(self, state_update: update_types.StateUpdate) -> None:
        if state_update.loaded_pipette != update_types.NO_CHANGE:
            pipette_id = state_update.loaded_pipette.pipette_id

            self._state.pipettes_by_id[pipette_id] = LoadedPipette(
                id=pipette_id,
                pipetteName=state_update.loaded_pipette.pipette_name,
                mount=state_update.loaded_pipette.mount,
            )
            self._state.liquid_presence_detection_by_id[pipette_id] = (
                state_update.loaded_pipette.liquid_presence_detection or False
            )
            self._state.aspirated_volume_by_id[pipette_id] = None
            self._state.movement_speed_by_id[pipette_id] = None
            self._state.attached_tip_by_id[pipette_id] = None

    def _update_tip_state(self, state_update: update_types.StateUpdate) -> None:
        if state_update.pipette_tip_state != update_types.NO_CHANGE:
            pipette_id = state_update.pipette_tip_state.pipette_id
            if state_update.pipette_tip_state.tip_geometry:
                attached_tip = state_update.pipette_tip_state.tip_geometry

                self._state.attached_tip_by_id[pipette_id] = attached_tip
                self._state.aspirated_volume_by_id[pipette_id] = 0

                static_config = self._state.static_config_by_id.get(pipette_id)
                if static_config:
                    try:
                        tip_configuration = (
                            static_config.tip_configuration_lookup_table[
                                attached_tip.volume
                            ]
                        )
                    except KeyError:
                        # TODO(seth,9/11/2023): this is a bad way of doing defaults but better than max volume.
                        # we used to look up a default tip config via the pipette max volume, but if that isn't
                        # tip volume (as it isn't when we're in low-volume mode) then that lookup fails. Using
                        # the first entry in the table is ok I guess but we really need to generally rethink how
                        # we identify tip classes - looking things up by volume is not enough.
                        tip_configuration = list(
                            static_config.tip_configuration_lookup_table.values()
                        )[0]
                    self._state.flow_rates_by_id[pipette_id] = FlowRates(
                        default_blow_out=tip_configuration.default_blowout_flowrate.values_by_api_level,
                        default_aspirate=tip_configuration.default_aspirate_flowrate.values_by_api_level,
                        default_dispense=tip_configuration.default_dispense_flowrate.values_by_api_level,
                    )

            else:
                pipette_id = state_update.pipette_tip_state.pipette_id
                self._state.aspirated_volume_by_id[pipette_id] = None
                self._state.attached_tip_by_id[pipette_id] = None

                static_config = self._state.static_config_by_id.get(pipette_id)
                if static_config:
                    # TODO(seth,9/11/2023): bad way to do defaulting, see above.
                    tip_configuration = list(
                        static_config.tip_configuration_lookup_table.values()
                    )[0]
                    self._state.flow_rates_by_id[pipette_id] = FlowRates(
                        default_blow_out=tip_configuration.default_blowout_flowrate.values_by_api_level,
                        default_aspirate=tip_configuration.default_aspirate_flowrate.values_by_api_level,
                        default_dispense=tip_configuration.default_dispense_flowrate.values_by_api_level,
                    )

    def _update_current_location(self, state_update: update_types.StateUpdate) -> None:
        location_update = state_update.pipette_location

        if location_update is update_types.NO_CHANGE:
            pass
        elif location_update is update_types.CLEAR:
            self._state.current_location = None
            self._state.current_deck_point = CurrentDeckPoint(
                mount=None, deck_point=None
            )
        else:
            new_logical_location = location_update.new_location
            new_deck_point = location_update.new_deck_point
            match new_logical_location:
                case update_types.Well(labware_id=labware_id, well_name=well_name):
                    self._state.current_location = CurrentWell(
                        pipette_id=location_update.pipette_id,
                        labware_id=labware_id,
                        well_name=well_name,
                    )
                case update_types.AddressableArea(
                    addressable_area_name=addressable_area_name
                ):
                    self._state.current_location = CurrentAddressableArea(
                        pipette_id=location_update.pipette_id,
                        addressable_area_name=addressable_area_name,
                    )
                case None:
                    self._state.current_location = None
                case update_types.NO_CHANGE:
                    pass
            if new_deck_point is not update_types.NO_CHANGE:
                loaded_pipette = self._state.pipettes_by_id[location_update.pipette_id]
                self._state.current_deck_point = CurrentDeckPoint(
                    mount=loaded_pipette.mount, deck_point=new_deck_point
                )

    def _update_pipette_config(self, state_update: update_types.StateUpdate) -> None:
        if state_update.pipette_config != update_types.NO_CHANGE:
            config = state_update.pipette_config.config
            self._state.static_config_by_id[
                state_update.pipette_config.pipette_id
            ] = StaticPipetteConfig(
                serial_number=state_update.pipette_config.serial_number,
                model=config.model,
                display_name=config.display_name,
                min_volume=config.min_volume,
                max_volume=config.max_volume,
                channels=config.channels,
                tip_configuration_lookup_table=config.tip_configuration_lookup_table,
                nominal_tip_overlap=config.nominal_tip_overlap,
                home_position=config.home_position,
                nozzle_offset_z=config.nozzle_offset_z,
                pipette_bounding_box_offsets=PipetteBoundingBoxOffsets(
                    back_left_corner=config.back_left_corner_offset,
                    front_right_corner=config.front_right_corner_offset,
                    back_right_corner=Point(
                        config.front_right_corner_offset.x,
                        config.back_left_corner_offset.y,
                        config.back_left_corner_offset.z,
                    ),
                    front_left_corner=Point(
                        config.back_left_corner_offset.x,
                        config.front_right_corner_offset.y,
                        config.back_left_corner_offset.z,
                    ),
                ),
                bounding_nozzle_offsets=BoundingNozzlesOffsets(
                    back_left_offset=config.nozzle_map.back_left_nozzle_offset,
                    front_right_offset=config.nozzle_map.front_right_nozzle_offset,
                ),
                default_nozzle_map=config.nozzle_map,
                lld_settings=config.pipette_lld_settings,
            )
            self._state.flow_rates_by_id[
                state_update.pipette_config.pipette_id
            ] = config.flow_rates
            self._state.nozzle_configuration_by_id[
                state_update.pipette_config.pipette_id
            ] = config.nozzle_map

    def _update_pipette_nozzle_map(
        self, state_update: update_types.StateUpdate
    ) -> None:
        if state_update.pipette_nozzle_map != update_types.NO_CHANGE:
            self._state.nozzle_configuration_by_id[
                state_update.pipette_nozzle_map.pipette_id
            ] = state_update.pipette_nozzle_map.nozzle_map

    def _update_volumes(
        self, action: Union[SucceedCommandAction, FailCommandAction]
    ) -> None:
        # todo(mm, 2024-10-10): Port these isinstance checks to StateUpdate.
        # https://opentrons.atlassian.net/browse/EXEC-754

        if isinstance(action, SucceedCommandAction) and isinstance(
            action.command.result,
            (commands.AspirateResult, commands.AspirateInPlaceResult),
        ):
            pipette_id = action.command.params.pipetteId
            previous_volume = self._state.aspirated_volume_by_id[pipette_id] or 0
            # PipetteHandler will have clamped action.command.result.volume for us, so
            # next_volume should always be in bounds.
            next_volume = previous_volume + action.command.result.volume

            self._state.aspirated_volume_by_id[pipette_id] = next_volume

        elif isinstance(action, SucceedCommandAction) and isinstance(
            action.command.result,
            (commands.DispenseResult, commands.DispenseInPlaceResult),
        ):
            pipette_id = action.command.params.pipetteId
            previous_volume = self._state.aspirated_volume_by_id[pipette_id] or 0
            # PipetteHandler will have clamped action.command.result.volume for us, so
            # next_volume should always be in bounds.
            next_volume = previous_volume - action.command.result.volume
            self._state.aspirated_volume_by_id[pipette_id] = next_volume

        elif isinstance(action, SucceedCommandAction) and isinstance(
            action.command.result,
            (
                commands.BlowOutResult,
                commands.BlowOutInPlaceResult,
                commands.unsafe.UnsafeBlowOutInPlaceResult,
            ),
        ):
            pipette_id = action.command.params.pipetteId
            self._state.aspirated_volume_by_id[pipette_id] = None

        elif isinstance(action, SucceedCommandAction) and isinstance(
            action.command.result, commands.PrepareToAspirateResult
        ):
            pipette_id = action.command.params.pipetteId
            self._state.aspirated_volume_by_id[pipette_id] = 0


class PipetteView(HasState[PipetteState]):
    """Read-only view of computed pipettes state."""

    _state: PipetteState

    def __init__(self, state: PipetteState) -> None:
        """Initialize the view with its backing state value."""
        self._state = state

    def get(self, pipette_id: str) -> LoadedPipette:
        """Get pipette data by the pipette's unique identifier."""
        try:
            return self._state.pipettes_by_id[pipette_id]
        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found."
            ) from e

    def get_mount(self, pipette_id: str) -> MountType:
        """Get the pipette's mount."""
        return self.get(pipette_id).mount

    def get_all(self) -> List[LoadedPipette]:
        """Get a list of all pipette entries in state."""
        return list(self._state.pipettes_by_id.values())

    def get_by_mount(self, mount: MountType) -> Optional[LoadedPipette]:
        """Get pipette data by the pipette's mount."""
        for pipette in self._state.pipettes_by_id.values():
            if pipette.mount == mount:
                return pipette
        return None

    def get_hardware_pipette(
        self,
        pipette_id: str,
        attached_pipettes: Mapping[HwMount, Optional[PipetteDict]],
    ) -> HardwarePipette:
        """Get a pipette's hardware configuration and state by ID."""
        pipette_data = self.get(pipette_id)
        pipette_name = pipette_data.pipetteName
        mount = pipette_data.mount

        hw_mount = mount.to_hw_mount()
        hw_config = attached_pipettes[hw_mount]

        # TODO(mc, 2022-01-11): HW controller may return an empty dict for
        # no pipette attached instead of `None`. Update when fixed in HWAPI
        if not hw_config:
            raise errors.PipetteNotAttachedError(f"No pipette attached on {mount}")

        elif (
            hw_config["name"] != pipette_name
            and pipette_name not in hw_config["back_compat_names"]
        ):
            raise errors.PipetteNotAttachedError(
                f"Found {hw_config['name']} on {mount}, "
                f"but {pipette_id} is a {pipette_name}"
            )

        return HardwarePipette(mount=hw_mount, config=hw_config)

    def get_current_location(self) -> Optional[CurrentPipetteLocation]:
        """Get the last accessed location and which pipette accessed it."""
        return self._state.current_location

    def get_deck_point(self, pipette_id: str) -> Optional[DeckPoint]:
        """Get the deck point of a pipette by ID, or None if it was not associated with the last move operation."""
        loaded_pipette = self.get(pipette_id)
        current_deck_point = self._state.current_deck_point
        if loaded_pipette.mount == current_deck_point.mount:
            return current_deck_point.deck_point
        return None

    def get_attached_tip(self, pipette_id: str) -> Optional[TipGeometry]:
        """Get details of the pipette's attached tip.

        Returns:
            The tip's volume and length, or None if there is no tip attached,
        """
        try:
            return self._state.attached_tip_by_id[pipette_id]
        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} no found; unable to get attached tip."
            ) from e

    def get_all_attached_tips(self) -> List[Tuple[str, TipGeometry]]:
        """Get a list of all attached tips.

        Returns:
            A list of pipette ID, tip details tuples.
        """
        return [
            (pipette_id, tip)
            for pipette_id, tip in self._state.attached_tip_by_id.items()
            if tip is not None
        ]

    def get_aspirated_volume(self, pipette_id: str) -> Optional[float]:
        """Get the currently aspirated volume of a pipette by ID.

        Returns:
            The volume the pipette has aspirated.
            None, after blow-out and the plunger is in an unsafe position.

        Raises:
            PipetteNotLoadedError: pipette ID does not exist.
            TipNotAttachedError: if no tip is attached to the pipette.
        """
        self.validate_tip_state(pipette_id, True)

        try:
            return self._state.aspirated_volume_by_id[pipette_id]

        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found; unable to get current volume."
            ) from e

    def get_working_volume(self, pipette_id: str) -> float:
        """Get the working maximum volume of a pipette by ID.

        Raises:
            PipetteNotLoadedError: pipette ID does not exist.
            TipNotAttachedError: if no tip is attached to the pipette.
        """
        max_volume = self.get_maximum_volume(pipette_id)
        attached_tip = self.get_attached_tip(pipette_id)

        if not attached_tip:
            raise errors.TipNotAttachedError(
                f"Pipette {pipette_id} has no tip attached; unable to calculate working maximum volume."
            )

        return min(attached_tip.volume, max_volume)

    def get_available_volume(self, pipette_id: str) -> Optional[float]:
        """Get the available volume of a pipette by ID."""
        working_volume = self.get_working_volume(pipette_id)
        current_volume = self.get_aspirated_volume(pipette_id)

        return max(0.0, working_volume - current_volume) if current_volume else None

    def get_pipette_lld_settings(
        self, pipette_id: str
    ) -> Optional[Dict[str, Dict[str, float]]]:
        """Get the liquid level settings for all possible tips for a single pipette."""
        return self.get_config(pipette_id).lld_settings

    def get_current_tip_lld_settings(self, pipette_id: str) -> float:
        """Get the liquid level settings for pipette and its current tip."""
        attached_tip = self.get_attached_tip(pipette_id)
        if attached_tip is None or attached_tip.volume is None:
            return 0
        lld_settings = self.get_pipette_lld_settings(pipette_id)
        tipVolume = "t" + str(int(attached_tip.volume))
        if (
            lld_settings is None
            or lld_settings[tipVolume] is None
            or lld_settings[tipVolume]["minHeight"] is None
        ):
            return 0
        return float(lld_settings[tipVolume]["minHeight"])

    def validate_tip_state(self, pipette_id: str, expected_has_tip: bool) -> None:
        """Validate that a pipette's tip state matches expectations."""
        attached_tip = self.get_attached_tip(pipette_id)

        if expected_has_tip is True and attached_tip is None:
            raise errors.TipNotAttachedError(
                "Pipette should have a tip attached, but does not."
            )
        if expected_has_tip is False and attached_tip is not None:
            raise errors.TipAttachedError(
                "Pipette should not have a tip attached, but does."
            )

    def get_movement_speed(
        self, pipette_id: str, requested_speed: Optional[float] = None
    ) -> Optional[float]:
        """Return the given pipette's requested or current movement speed."""
        return requested_speed or self._state.movement_speed_by_id[pipette_id]

    def get_config(self, pipette_id: str) -> StaticPipetteConfig:
        """Get the static pipette configuration by pipette id."""
        try:
            return self._state.static_config_by_id[pipette_id]
        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found; unable to get pipette configuration."
            ) from e

    def get_model_name(self, pipette_id: str) -> str:
        """Return the given pipette's model name."""
        return self.get_config(pipette_id).model

    def get_display_name(self, pipette_id: str) -> str:
        """Return the given pipette's display name."""
        return self.get_config(pipette_id).display_name

    def get_serial_number(self, pipette_id: str) -> str:
        """Get the serial number of the pipette."""
        return self.get_config(pipette_id).serial_number

    def get_channels(self, pipette_id: str) -> int:
        """Return the max channels of the pipette."""
        return self.get_config(pipette_id).channels

    def get_minimum_volume(self, pipette_id: str) -> float:
        """Return the given pipette's minimum volume."""
        return self.get_config(pipette_id).min_volume

    def get_maximum_volume(self, pipette_id: str) -> float:
        """Return the given pipette's maximum volume."""
        return self.get_config(pipette_id).max_volume

    def get_instrument_max_height_ot2(self, pipette_id: str) -> float:
        """Get calculated max instrument height for an OT-2."""
        config = self.get_config(pipette_id)
        return config.home_position - Z_RETRACT_DISTANCE + config.nozzle_offset_z

    def get_return_tip_scale(self, pipette_id: str) -> float:
        """Return the given pipette's return tip height scale."""
        max_volume = self.get_maximum_volume(pipette_id)
        working_volume = max_volume
        tip = self.get_attached_tip(pipette_id)
        if tip:
            working_volume = tip.volume

        if working_volume in self.get_config(pipette_id).tip_configuration_lookup_table:
            tip_lookup = self.get_config(pipette_id).tip_configuration_lookup_table[
                working_volume
            ]
        else:
            tip_lookup = self.get_config(pipette_id).tip_configuration_lookup_table[
                max_volume
            ]
        return tip_lookup.default_return_tip_height

    def get_flow_rates(self, pipette_id: str) -> FlowRates:
        """Get the default flow rates for the pipette."""
        try:
            return self._state.flow_rates_by_id[pipette_id]
        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found; unable to get pipette flow rates."
            ) from e

    def get_nominal_tip_overlap(self, pipette_id: str, labware_uri: str) -> float:
        """Get the nominal tip overlap for a given labware from config."""
        tip_overlaps_by_uri = self.get_config(pipette_id).nominal_tip_overlap

        try:
            return tip_overlaps_by_uri[labware_uri]
        except KeyError:
            return tip_overlaps_by_uri.get("default", 0)

    def get_z_axis(self, pipette_id: str) -> MotorAxis:
        """Get the MotorAxis representing this pipette's Z stage."""
        mount = self.get(pipette_id).mount
        return MotorAxis.LEFT_Z if mount == MountType.LEFT else MotorAxis.RIGHT_Z

    def get_plunger_axis(self, pipette_id: str) -> MotorAxis:
        """Get the MotorAxis representing this pipette's plunger."""
        mount = self.get(pipette_id).mount
        return (
            MotorAxis.LEFT_PLUNGER
            if mount == MountType.LEFT
            else MotorAxis.RIGHT_PLUNGER
        )

    def get_nozzle_layout_type(self, pipette_id: str) -> NozzleConfigurationType:
        """Get the current set nozzle layout configuration."""
        nozzle_map_for_pipette = self._state.nozzle_configuration_by_id[pipette_id]
        return nozzle_map_for_pipette.configuration

    def get_is_partially_configured(self, pipette_id: str) -> bool:
        """Determine if the provided pipette is partially configured."""
        return self.get_nozzle_layout_type(pipette_id) != NozzleConfigurationType.FULL

    def get_primary_nozzle(self, pipette_id: str) -> str:
        """Get the primary nozzle, if any, related to the given pipette's nozzle configuration."""
        nozzle_map = self._state.nozzle_configuration_by_id[pipette_id]
        return nozzle_map.starting_nozzle

    def _get_critical_point_offset_without_tip(
        self, pipette_id: str, critical_point: Optional[CriticalPoint]
    ) -> Point:
        """Get the offset of the specified critical point from pipette's mount position."""
        nozzle_map = self._state.nozzle_configuration_by_id[pipette_id]
        match critical_point:
            case CriticalPoint.INSTRUMENT_XY_CENTER:
                return nozzle_map.instrument_xy_center_offset
            case CriticalPoint.XY_CENTER:
                return nozzle_map.xy_center_offset
            case CriticalPoint.Y_CENTER:
                return nozzle_map.y_center_offset
            case CriticalPoint.FRONT_NOZZLE:
                return nozzle_map.front_nozzle_offset
            case _:
                return nozzle_map.starting_nozzle_offset

    def get_pipette_bounding_nozzle_offsets(
        self, pipette_id: str
    ) -> BoundingNozzlesOffsets:
        """Get the nozzle offsets of the pipette's bounding nozzles."""
        return self.get_config(pipette_id).bounding_nozzle_offsets

    def get_pipette_bounding_box(self, pipette_id: str) -> PipetteBoundingBoxOffsets:
        """Get the bounding box of the pipette."""
        return self.get_config(pipette_id).pipette_bounding_box_offsets

    # TODO (spp, 2024-09-17): in order to find the position of pipette at destination,
    #  this method repeats the same steps that waypoints builder does while finding
    #  waypoints to move to. We should consolidate these steps into a shared entity
    #  so that the deck conflict checker and movement plan builder always remain in sync.
    def get_pipette_bounds_at_specified_move_to_position(
        self,
        pipette_id: str,
        destination_position: Point,
        critical_point: Optional[CriticalPoint],
    ) -> Tuple[Point, Point, Point, Point]:
        """Get the pipette's bounding box position when critical point is at the destination position.

        Returns a tuple of the pipette's bounding box position in deck coordinates as-
            (back_left_bound, front_right_bound, back_right_bound, front_left_bound)
        Bounding box of the pipette includes the pipette's outer casing as well as nozzles.
        """
        tip = self.get_attached_tip(pipette_id)

        # *Offset* of pipette's critical point w.r.t pipette mount
        critical_point_offset = self._get_critical_point_offset_without_tip(
            pipette_id, critical_point
        )

        # Position of the above critical point at destination, in deck coordinates
        critical_point_position = destination_position + Point(
            x=0, y=0, z=tip.length if tip else 0
        )

        # Get the pipette bounding box coordinates
        pipette_bounds_offsets = self.get_config(
            pipette_id
        ).pipette_bounding_box_offsets
        pip_back_left_bound = (
            critical_point_position
            - critical_point_offset
            + pipette_bounds_offsets.back_left_corner
        )
        pip_front_right_bound = (
            critical_point_position
            - critical_point_offset
            + pipette_bounds_offsets.front_right_corner
        )
        pip_back_right_bound = Point(
            pip_front_right_bound.x, pip_back_left_bound.y, pip_front_right_bound.z
        )
        pip_front_left_bound = Point(
            pip_back_left_bound.x, pip_front_right_bound.y, pip_back_left_bound.z
        )
        return (
            pip_back_left_bound,
            pip_front_right_bound,
            pip_back_right_bound,
            pip_front_left_bound,
        )

    def get_liquid_presence_detection(self, pipette_id: str) -> bool:
        """Determine if liquid presence detection is enabled for this pipette."""
        try:
            return self._state.liquid_presence_detection_by_id[pipette_id]
        except KeyError as e:
            raise errors.PipetteNotLoadedError(
                f"Pipette {pipette_id} not found; unable to determine if pipette liquid presence detection enabled."
            ) from e
